# if someone hits a 404, redirect them to another location


def send_http_302_temporary_redirect(cli, new_path):
    """
    replies with an HTTP 302, which is a temporary redirect;
    "new_path" can be any of the following:
      - "http://a.com/" would redirect to another website,
      - "/foo/bar" would redirect to /foo/bar on the same server;
          note the leading '/' in the location which is important
    """
    cli.reply(b"redirecting...", 302, headers={"Location": new_path})


def send_http_301_permanent_redirect(cli, new_path):
    """
    replies with an HTTP 301, which is a permanent redirect;
    otherwise identical to send_http_302_temporary_redirect
    """
    cli.reply(b"redirecting...", 301, headers={"Location": new_path})


def send_errorpage_with_redirect_link(cli, new_path):
    """
    replies with a website explaining that the page has moved;
    "new_path" must be an absolute location on the same server
    but without a leading '/', so for example "foo/bar"
    would redirect to "/foo/bar"
    """
    cli.redirect(new_path, click=False, msg="this page has moved")


def main(cli, vn, rem):
    """
    this is the function that gets called by copyparty;
    note that vn.vpath and cli.vpath does not have a leading '/'
    so we're adding the slash in the debug messages below
    """
    print(f"this client just hit a 404: {cli.ip}")
    print(f"they were accessing this volume: /{vn.vpath}")
    print(f"and the original request-path (straight from the URL) was /{cli.vpath}")
    print(f"...which resolves to the following filesystem path: {vn.canonical(rem)}")

    new_path = "/foo/bar/"
    print(f"will now redirect the client to {new_path}")

    # uncomment one of these:
    send_http_302_temporary_redirect(cli, new_path)
    # send_http_301_permanent_redirect(cli, new_path)
    # send_errorpage_with_redirect_link(cli, new_path)

    return "true"



#!/usr/bin/env python3

import json
import re
import os
import sys
import subprocess as sp

try:
    from copyparty.util import fsenc
except:

    def fsenc(p):
        return p.encode("utf-8")


_ = r"""
inspects video files for errors and such
plus stores a bunch of metadata to filename.ff.json

usage:
  -mtp vidchk=t600,ay,p,bin/mtag/vidchk.py

explained:
t600: timeout 10min
  ay: only process files which contain audio (including video with audio)
   p: set priority 1 (lowest priority after initial ffprobe/mutagen for base tags),
       makes copyparty feed base tags into this script as json

if you wanna use this script standalone / separately from copyparty,
provide the video resolution on stdin as json:  {"res":"1920x1080"}
"""


FAST = True  # parse entire file at container level
# FAST = False  # fully decode audio and video streams


# warnings to ignore
harmless = re.compile(
    r"Unsupported codec with id |Could not find codec parameters.*Attachment:|analyzeduration"
    + r"|timescale not set"
)


def wfilter(lines):
    return [x for x in lines if x.strip() and not harmless.search(x)]


def errchk(so, se, rc, dbg):
    if dbg:
        with open(dbg, "wb") as f:
            f.write(b"so:\n" + so + b"\nse:\n" + se + b"\n")

    if rc:
        err = (so + se).decode("utf-8", "replace").split("\n", 1)
        err = wfilter(err) or err
        return f"ERROR {rc}: {err[0]}"

    if se:
        err = se.decode("utf-8", "replace").split("\n", 1)
        err = wfilter(err)
        if err:
            return f"Warning: {err[0]}"

    return None


def main():
    fp = sys.argv[1]
    zb = sys.stdin.buffer.read()
    zs = zb.decode("utf-8", "replace")
    md = json.loads(zs)

    fdir = os.path.dirname(os.path.realpath(fp))
    flag = os.path.join(fdir, ".processed")
    if os.path.exists(flag):
        return "already processed"

    try:
        w, h = [int(x) for x in md["res"].split("x")]
        if not w + h:
            raise Exception()
    except:
        return "could not determine resolution"

    # grab streams/format metadata + 2 seconds of frames at the start and end
    zs = "ffprobe -hide_banner -v warning -of json -show_streams -show_format -show_packets -show_data_hash crc32 -read_intervals %+2,999999%+2"
    cmd = zs.encode("ascii").split(b" ") + [fsenc(fp)]
    p = sp.Popen(cmd, stdout=sp.PIPE, stderr=sp.PIPE)
    so, se = p.communicate()

    # spaces to tabs, drops filesize from 69k to 48k
    so = b"\n".join(
        [
            b"\t" * int((len(x) - len(x.lstrip())) / 4) + x.lstrip()
            for x in (so or b"").split(b"\n")
        ]
    )
    with open(fsenc(f"{fp}.ff.json"), "wb") as f:
        f.write(so)

    err = errchk(so, se, p.returncode, f"{fp}.vidchk")
    if err:
        return err

    if max(w, h) < 1280 and min(w, h) < 720:
        return "resolution too small"

    zs = (
        "ffmpeg -y -hide_banner -nostdin -v warning"
        + " -err_detect +crccheck+bitstream+buffer+careful+compliant+aggressive+explode"
        + " -xerror -i"
    )

    cmd = zs.encode("ascii").split(b" ") + [fsenc(fp)]

    if FAST:
        zs = "-c copy -f null -"
    else:
        zs = "-vcodec rawvideo -acodec pcm_s16le -f null -"

    cmd += zs.encode("ascii").split(b" ")

    p = sp.Popen(cmd, stdout=sp.PIPE, stderr=sp.PIPE)
    so, se = p.communicate()
    return errchk(so, se, p.returncode, f"{fp}.vidchk")


if __name__ == "__main__":
    print(main() or "ok")



#!/usr/bin/env python

import os
import sys
import tempfile
import subprocess as sp

try:
    import keyfinder

    PKF = True
except:
    PKF = False

from copyparty.util import fsenc

"""
dep: github/mixxxdj/libkeyfinder
dep: pypi/keyfinder  -OR-  EvanPurkhiser/keyfinder-cli
dep: ffmpeg
"""


# tried trimming the first/last 5th, bad idea,
# misdetects 9a law field (Sphere Caliber) as 10b,
# obvious when mixing 9a ghostly parapara ship


def det(tf):
    # fmt: off
    sp.check_call([
        b"ffmpeg",
        b"-nostdin",
        b"-hide_banner",
        b"-v", b"fatal",
        b"-y", b"-i", fsenc(sys.argv[1]),
        b"-map", b"0:a:0",
        b"-t", b"300",
        b"-sample_fmt", b"s16",
        fsenc(tf)
    ])
    # fmt: on

    if PKF:
        print(keyfinder.key(tf).camelot())
    else:
        # fmt: off
        sp.check_call([
            b"keyfinder-cli",
            b"-n",
            b"camelot",
            fsenc(tf)
        ])
        # fmt: on


def main():
    with tempfile.NamedTemporaryFile(suffix=".flac", delete=False) as f:
        f.write(b"h")
        tf = f.name

    try:
        det(tf)
    except:
        pass  # mute
    finally:
        os.unlink(tf)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# coding: utf-8
from __future__ import print_function, unicode_literals

import os
import shutil
import tempfile
import unittest

from copyparty.authsrv import AuthSrv
from copyparty.httpcli import HttpCli
from tests import util as tu
from tests.util import Cfg

try:
    from typing import Optional
except:
    pass


def hdr(query):
    h = "GET /{} HTTP/1.1\r\nPW: o\r\nConnection: close\r\n\r\n"
    return h.format(query).encode("utf-8")


class TestHooks(tu.TC):
    def setUp(self):
        self.conn: Optional[tu.VHttpConn] = None
        self.td = tu.get_ramdisk()

    def tearDown(self):
        os.chdir(tempfile.gettempdir())
        shutil.rmtree(self.td)

    def reset(self):
        td = os.path.join(self.td, "vfs")
        if os.path.exists(td):
            shutil.rmtree(td)
        os.mkdir(td)
        os.chdir(td)
        return td

    def cinit(self):
        if self.conn:
            self.conn.shutdown()
            self.conn = None
        self.conn = tu.VHttpConn(self.args, self.asrv, self.log, b"")

    def test(self):
        vcfg = ["a/b/c/d:c/d:A", "a:a:r"]

        scenarios = (
            ('{"vp":"x/y"}', "c/d/a.png", "c/d/x/y/a.png"),
            ('{"vp":"x/y"}', "c/d/e/a.png", "c/d/e/x/y/a.png"),
            ('{"vp":"../x/y"}', "c/d/e/a.png", "c/d/x/y/a.png"),
            ('{"ap":"x/y"}', "c/d/a.png", "c/d/x/y/a.png"),
            ('{"ap":"x/y"}', "c/d/e/a.png", "c/d/e/x/y/a.png"),
            ('{"ap":"../x/y"}', "c/d/e/a.png", "c/d/x/y/a.png"),
            ('{"ap":"../x/y"}', "c/d/a.png", "a/b/c/x/y/a.png"),
            ('{"fn":"b.png"}', "c/d/a.png", "c/d/b.png"),
            ('{"vp":"x","fn":"b.png"}', "c/d/a.png", "c/d/x/b.png"),
        )

        for x in scenarios:
            print("\n\n\n", x)
            hooktxt, url_up, url_dl = x
            for hooktype in ("xbu", "xau"):
                for upfun in (self.put, self.bup):
                    self.reset()
                    self.makehook("""print('{"reloc":%s}')""" % (hooktxt,))
                    ka = {hooktype: ["j,c1,h.py"]}
                    self.args = Cfg(v=vcfg, a=["o:o"], e2d=True, **ka)
                    self.asrv = AuthSrv(self.args, self.log)
                    self.cinit()

                    h, b = upfun(url_up)
                    self.assertStart("HTTP/1.1 201 Created\r", h)
                    h, b = self.curl(url_dl)
                    self.assertEqual(b, "ok %s\n" % (url_up))

    def test2(self):
        hooktxt = "import sys\nopen('h%d','wb').close()\nsys.exit(%d)\n"
        for hooktype in ("xbu", "xau"):
            for upfun in (self.put, self.bup):
                self.reset()
                for n in [0, 1, 100]:
                    with open("h%d.py" % (n,), "wb") as f:
                        f.write((hooktxt % (n, n)).encode("utf-8"))
                vcfg = [
                    "012:012:A:c,H=c,h0.py:c,H=c,h1.py:c,H=c,h100.py",
                    "021:021:A:c,H=c,h0.py:c,H=c,h100.py:c,H=c,h1.py",
                    "120:120:A:c,H=c,h1.py:c,H=c,h100.py:c,H=c,h0.py",
                    "30:30:A:c,H=c,enoent.py:c,H=c,h100.py",  # not-exist
                ]
                vcfg = [x.replace("H", hooktype) for x in vcfg]
                self.args = Cfg(v=vcfg, a=["o:o"], e2d=True)
                self.asrv = AuthSrv(self.args, self.log)
                self.cinit()
                scenarios = (
                    ("012", False, True, True, False),
                    ("021", True, True, False, True),
                    ("120", False, False, True, False),
                    ("30", False, False, False, False),
                )
                for (vp, ok, h0, h1, h2) in scenarios:
                    for zs in ("h0", "h1", "h100"):
                        if os.path.exists(zs):
                            os.unlink(zs)
                    vp = "%s/f" % (vp,)
                    h, b = upfun(vp)
                    self.assertEqual(ok, os.path.exists(vp))
                    self.assertEqual(h0, os.path.exists("h0"))
                    self.assertEqual(h1, os.path.exists("h1"))
                    self.assertEqual(h2, os.path.exists("h100"))

    def makehook(self, hs):
        with open("h.py", "wb") as f:
            f.write(hs.encode("utf-8"))

    def put(self, url):
        buf = "PUT /{0} HTTP/1.1\r\nPW: o\r\nConnection: close\r\nContent-Length: {1}\r\n\r\nok {0}\n"
        buf = buf.format(url, len(url) + 4).encode("utf-8")
        print("PUT -->", buf)
        conn = self.conn.setbuf(buf)
        HttpCli(conn).run()
        ret = conn.s._reply.decode("utf-8").split("\r\n\r\n", 1)
        print("PUT <--", ret)
        return ret

    def bup(self, url):
        hdr = "POST /%s HTTP/1.1\r\nPW: o\r\nConnection: close\r\nContent-Type: multipart/form-data; boundary=XD\r\nContent-Length: %d\r\n\r\n"
        bdy = '--XD\r\nContent-Disposition: form-data; name="act"\r\n\r\nbput\r\n--XD\r\nContent-Disposition: form-data; name="f"; filename="%s"\r\n\r\n'
        ftr = "\r\n--XD--\r\n"
        try:
            url, fn = url.rsplit("/", 1)
        except:
            fn = url
            url = ""

        buf = (bdy % (fn,) + "ok %s/%s\n" % (url, fn) + ftr).encode("utf-8")
        buf = (hdr % (url, len(buf))).encode("utf-8") + buf
        print("PoST -->", buf)
        conn = self.conn.setbuf(buf)
        HttpCli(conn).run()
        ret = conn.s._reply.decode("utf-8").split("\r\n\r\n", 1)
        print("POST <--", ret)
        return ret

    def curl(self, url, binary=False):
        conn = self.conn.setbuf(hdr(url))
        HttpCli(conn).run()
        if binary:
            h, b = conn.s._reply.split(b"\r\n\r\n", 1)
            return [h.decode("utf-8"), b]

        return conn.s._reply.decode("utf-8").split("\r\n\r\n", 1)

    def log(self, src, msg, c=0):
        print(msg)



#!/usr/bin/env python3

import os
import sys
import subprocess as sp
from plyer import notification


_ = r"""
show os notification on upload; works on windows, linux, macos, android

dependencies:
    windows: python3 -m pip install --user -U plyer
    linux:   python3 -m pip install --user -U plyer
    macos:   python3 -m pip install --user -U plyer pyobjus
    android: just termux and termux-api

example usages; either as global config (all volumes) or as volflag:
    --xau f,bin/hooks/notify.py
    -v srv/inc:inc:r:rw,ed:c,xau=f,bin/hooks/notify.py
                           ^^^^^^^^^^^^^^^^^^^^^^^^^^^

    (share filesystem-path srv/inc as volume /inc,
     readable by everyone, read-write for user 'ed',
     running this plugin on all uploads with the params listed below)

parameters explained,
    xau = execute after upload
    f   = fork so it doesn't block uploads
"""


try:
    from copyparty.util import humansize
except:

    def humansize(n):
        return n


def main():
    fp = sys.argv[1]
    dp, fn = os.path.split(fp)
    try:
        sz = humansize(os.path.getsize(fp))
    except:
        sz = "?"

    msg = "{} ({})\n📁 {}".format(fn, sz, dp)
    title = "File received"

    if "com.termux" in sys.executable:
        sp.run(["termux-notification", "-t", title, "-c", msg])
        return

    icon = "emblem-documents-symbolic" if sys.platform == "linux" else ""
    notification.notify(
        title=title,
        message=msg,
        app_icon=icon,
        timeout=10,
    )


if __name__ == "__main__":
    main()



# create a dummy file and let copyparty return it


def main(cli, vn, rem):
    print("hello", cli.ip)

    abspath = vn.canonical(rem)
    with open(abspath, "wb") as f:
        f.write(b"404? not on MY watch!")

    return "retry"



#!/usr/bin/env python3

import os
import sys


_ = r"""
rename incoming uploads according to the "wark" (the file identifier)
which is basically but not exactly a sha512 hash of the file contents

NOTE: this does NOT work with up2k uploads (dragdrop into browser);
  combine this hook with reloc-by-wark-xbu.py to fix that

example usage as global config:
    -e2d --xau I,c,bin/hooks/reloc-by-wark-xau.py

parameters explained,
    e2d = enable up2k database (mandatory for xau hooks)
    xau = execute before upload
    I   = import this hook for performance; do not fork / subprocess
    c   = "check"; reject upload if this hook crashes due to a bug

example usage as a volflag (per-volume config):
    -v srv/inc:inc:r:rw,ed:c,e2d,xau=I,c,bin/hooks/reloc-by-wark-xau.py
                           ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

    (share filesystem-path srv/inc as volume /inc,
     readable by everyone, read-write for user 'ed',
     running this plugin on all uploads with the params explained above)

example usage as a volflag in a copyparty config file:
    [/inc]
      srv/inc
      accs:
        r: *
        rw: ed
      flags:
        e2d, xau: I,c,bin/hooks/reloc-by-wark-xau.py
"""


def main(inf):
    if inf.get("wark"):
        # this is an up2k upload; nothing to be done, just bail
        return {}

    abspath = inf["ap"]  # filesystem path to the uploaded file

    # we don't have the wark yet so need to calculate it;
    # generating a regular sha512 would of course be much easier,
    # but then filenames would be different depending on how the
    # file was uploaded (laaame) so let's do it the hard way

    # use the standard up2k-salt which nobody ever changes:
    salt = "hunter2"

    # to generate the wark we'll need some functions from copyparty;
    # follow the trail to the copyparty module and grab them from there:

    import inspect

    libpath = inspect.getfile(inf["log"])
    libpath = os.path.dirname(os.path.dirname(libpath))
    sys.path.insert(0, libpath)

    from copyparty.up2k import up2k_hashlist_from_file, up2k_wark_from_hashlist

    chunklist, st = up2k_hashlist_from_file(abspath)
    wark = up2k_wark_from_hashlist(salt, st.st_size, chunklist)

    # okay nice
    # the rest of the code below is just copied from reloc-by-wark-xbu.py
    # -------------------------------------------------------------------------

    # grab the original filename from the vpath...
    vdir, fn = os.path.split(inf["vp"])

    # ...to retain the original file extension, if any
    try:
        fn, ext = fn.rsplit(".", 1)
    except:
        ext = ""

    # use the first 16 characters; 12 bytes of entropy,
    # roughly one collision for every 26 million files
    fn = wark[:16]

    if ext:
        ext = ext.lower()
        fn += "." + ext

    return {"reloc": {"fn": fn}}



import json
import re
import sys

from copyparty.util import fsenc, runcmd


"""
uses exiftool to geotag images based on embedded gps coordinates in exif data

adds four new metadata keys:
    .gps_lat = latitute
    .gps_lon = longitude
    .masl = meters above sea level
    city = "city, subregion, region"

usage: -mtp .masl,.gps_lat,.gps_lon,city=ad,t10,bin/mtag/geotag.py

example: https://a.ocv.me/pub/blog/j7/8/?grid=0
"""


def main():
    cmd = b"exiftool -api geolocation -n".split(b" ")
    rc, so, se = runcmd(cmd + [fsenc(sys.argv[1])])
    ptn = re.compile("([^:]*[^ :]) *: (.*)")
    city = ["", "", ""]
    ret = {}
    for ln in so.split("\n"):
        m = ptn.match(ln)
        if not m:
            continue
        k, v = m.groups()
        if k == "Geolocation City":
            city[2] = v
        elif k == "Geolocation Subregion":
            city[1] = v
        elif k == "Geolocation Region":
            city[0] = v
        elif k == "GPS Latitude":
            ret[".gps_lat"] = "%.04f" % (float(v),)
        elif k == "GPS Longitude":
            ret[".gps_lon"] = "%.04f" % (float(v),)
        elif k == "GPS Altitude":
            ret[".masl"] = str(int(float(v)))
    v = ", ".join(city).strip(", ")
    if v:
        ret["city"] = v
    print(json.dumps(ret))


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# coding: utf-8

import os
import sys
import json
import shutil
import subprocess as sp


_ = r"""
start downloading a torrent by POSTing a magnet URL to copyparty,
for example using 📟 (message-to-server-log) in the web-ui

by default it will download the torrent to the folder you were in
when you pasted the magnet into the message-to-server-log field

you can optionally specify another location by adding a whitespace
after the magnet URL followed by the name of the subfolder to DL into,
or for example "anime/airing" would download to /srv/media/anime/airing
because the keyword "anime" is in the DESTS config below

needs python3

example usage as global config (not a good idea):
    python copyparty-sfx.py --xm aw,f,j,t60,bin/hooks/qbittorrent-magnet.py

parameters explained,
    xm = execute on message (📟)
    aw = only users with write-access can use this
    f = fork; don't delay other hooks while this is running
    j = provide message information as json (not just the text)
    t60 = abort if qbittorrent has to think about it for more than 1 min

example usage as a volflag (per-volume config, much better):
    -v srv/qb:qb:A,ed:c,xm=aw,f,j,t60,bin/hooks/qbittorrent-magnet.py
                      ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

    (share filesystem-path srv/qb as volume /qb with Admin for user 'ed',
     running this plugin on all messages with the params explained above)

example usage as a volflag in a copyparty config file:
    [/qb]
      srv/qb
      accs:
        A: ed
      flags:
        xm: aw,f,j,t60,bin/hooks/qbittorrent-magnet.py

the volflag examples only kicks in if you send the torrent magnet
while you're in the /qb folder (or any folder below there)
"""


# list of usernames to allow
ALLOWLIST = [ "ed", "morpheus" ]


# list of destination aliases to translate into full filesystem
# paths; takes effect if the first folder component in the
# custom download location matches anything in this dict
DESTS = {
    "iso": "/srv/pub/linux-isos",
    "anime": "/srv/media/anime",
}


def main():
    inf = json.loads(sys.argv[1])
    url = inf["txt"]
    if not url.lower().startswith("magnet:?"):
        # not a magnet, abort
        return

    if inf["user"] not in ALLOWLIST:
        print("🧲 denied for user", inf["user"])
        return

    # might as well run the command inside the filesystem folder
    # which matches the URL that the magnet message was sent to
    os.chdir(inf["ap"])

    # is there is a custom download location in the url?
    dst = ""
    if " " in url:
        url, dst = url.split(" ", 1)

        # is the location in the predefined list of locations?
        parts = dst.replace("\\", "/").split("/")
        if parts[0] in DESTS:
            dst = os.path.join(DESTS[parts[0]], *(parts[1:]))

    else:
        # nope, so download to the current folder instead;
        # comment the dst line below to instead use the default
        # download location from your qbittorrent settings
        dst = inf["ap"]
        pass

    # archlinux has a -nox suffix for qbittorrent if headless
    # so check if we should be using that
    if shutil.which("qbittorrent-nox"):
        torrent_bin = "qbittorrent-nox"
    else:
        torrent_bin = "qbittorrent"

    # the command to add a new torrent, adjust if necessary
    cmd = [torrent_bin, url]
    if dst:
        cmd += ["--save-path=%s" % (dst,)]

    # if copyparty and qbittorrent are running as different users
    # you may have to do something like the following
    # (assuming qbittorrent* is nopasswd-allowed in sudoers):
    #
    # cmd = ["sudo", "-u", "qbitter"] + cmd

    print("🧲", cmd)

    try:
        sp.check_call(cmd)
    except:
        print("🧲 FAILED TO ADD", url)


if __name__ == "__main__":
    main()




#!/usr/bin/env python3

import sys


_ = r"""
reject file uploads by file extension

example usage as global config:
    --xbu c,bin/hooks/reject-extension.py

example usage as a volflag (per-volume config):
    -v srv/inc:inc:r:rw,ed:c,xbu=c,bin/hooks/reject-extension.py
                           ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

    (share filesystem-path srv/inc as volume /inc,
     readable by everyone, read-write for user 'ed',
     running this plugin on all uploads with the params listed below)

parameters explained,
    xbu = execute before upload
    c   = check result, reject upload if error
"""


def main():
    bad = "exe scr com pif bat ps1 jar msi"

    ext = sys.argv[1].split(".")[-1]

    sys.exit(1 if ext in bad.split() else 0)


if __name__ == "__main__":
    main()



#!/usr/bin/env python

import json
import os
import subprocess as sp
import sys
import time

try:
    from copyparty.util import fsenc
except:

    def fsenc(p):
        return p.encode("utf-8")


_ = r"""
first checks the tag "vidchk" which must be "ok" to continue,
then uploads all files to some cloud storage (RCLONE_REMOTE)
and DELETES THE ORIGINAL FILES if rclone returns 0 ("success")

deps:
  rclone

usage:
  -mtp x2=t43200,ay,p2,bin/mtag/rclone-upload.py

explained:
t43200: timeout 12h
    ay: only process files which contain audio (including video with audio)
    p2: set priority 2 (after vidchk's suggested priority of 1),
          so the output of vidchk will be passed in here

complete usage example as vflags along with vidchk:
  -vsrv/vidchk:vidchk:r:rw,ed:c,e2dsa,e2ts,mtp=vidchk=t600,p,bin/mtag/vidchk.py:c,mtp=rupload=t43200,ay,p2,bin/mtag/rclone-upload.py:c,mte=+vidchk,rupload

setup: see https://rclone.org/drive/

if you wanna use this script standalone / separately from copyparty,
either set CONDITIONAL_UPLOAD False or provide the following stdin:
  {"vidchk":"ok"}
"""


RCLONE_REMOTE = "notmybox"
CONDITIONAL_UPLOAD = True


def main():
    fp = sys.argv[1]
    if CONDITIONAL_UPLOAD:
        zb = sys.stdin.buffer.read()
        zs = zb.decode("utf-8", "replace")
        md = json.loads(zs)

        chk = md.get("vidchk", None)
        if chk != "ok":
            print(f"vidchk={chk}", file=sys.stderr)
            sys.exit(1)

    dst = f"{RCLONE_REMOTE}:".encode("utf-8")
    cmd = [b"rclone", b"copy", b"--", fsenc(fp), dst]

    t0 = time.time()
    try:
        sp.check_call(cmd)
    except:
        print("rclone failed", file=sys.stderr)
        sys.exit(1)

    print(f"{time.time() - t0:.1f} sec")
    os.unlink(fsenc(fp))


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# coding: utf-8
from __future__ import print_function, unicode_literals

import re
import unittest
from xml.etree import ElementTree as ET

from copyparty.dxml import BadXML, mkenod, mktnod, parse_xml

ET.register_namespace("D", "DAV:")


def _parse(txt):
    try:
        parse_xml(txt)
        raise Exception("unsafe")
    except BadXML:
        pass


class TestDXML(unittest.TestCase):
    def test_qbe(self):
        # allowed by default; verify that we stopped it
        txt = r"""<!DOCTYPE qbe [
<!ENTITY a "nice_bakuretsu">
]>
<l>&a;&a;&a;&a;&a;&a;&a;&a;&a;</l>"""
        _parse(txt)
        ET.fromstring(txt)

    def test_ent_file(self):
        # NOT allowed by default; should still be blocked
        txt = r"""<!DOCTYPE ext [
<!ENTITY ee SYSTEM "file:///bin/bash">
]>
<root>&ee;</root>"""
        _parse(txt)
        try:
            ET.fromstring(txt)
            raise Exception("unsafe2")
        except ET.ParseError:
            pass

    def test_ent_ext(self):
        # NOT allowed by default; should still be blocked
        txt = r"""<!DOCTYPE ext [
<!ENTITY ee SYSTEM "http://example.com/a.xml">
]>
<root>&ee;</root>"""
        _parse(txt)

    def test_dtd(self):
        # allowed by default; verify that we stopped it
        txt = r"""<!DOCTYPE d SYSTEM "a.dtd">
<root>a</root>"""
        _parse(txt)
        ET.fromstring(txt)

    ##
    ## end of negative/security tests; the rest is functional
    ##

    def test3(self):
        txt = r"""<?xml version="1.0" ?>
<propfind xmlns="DAV:">
    <prop>
        <name/>
        <href/>
    </prop>
</propfind>
"""
        txt = txt.replace("\n", "\r\n")
        ET.fromstring(txt)
        el = parse_xml(txt)
        self.assertListEqual(
            [y.tag for y in el.findall(r"./{DAV:}prop/*")],
            [r"{DAV:}name", r"{DAV:}href"],
        )

    def test4(self):
        txt = r"""<?xml version="1.0" encoding="utf-8" ?>
<D:propertyupdate xmlns:D="DAV:" xmlns:Z="urn:schemas-microsoft-com:">
    <D:set>
        <D:prop>
            <Z:Win32CreationTime>Thu, 20 Oct 2022 02:16:33 GMT</Z:Win32CreationTime>
            <Z:Win32LastAccessTime>Thu, 20 Oct 2022 02:16:35 GMT</Z:Win32LastAccessTime>
            <Z:Win32LastModifiedTime>Thu, 20 Oct 2022 02:16:33 GMT</Z:Win32LastModifiedTime>
            <Z:Win32FileAttributes>00000000</Z:Win32FileAttributes>
        </D:prop>
    </D:set>
</D:propertyupdate>"""

        ref = r"""<?xml version="1.0" encoding="UTF-8"?>
<D:multistatus xmlns:D="DAV:">
    <D:response>
        <D:href>/d1/foo.txt</D:href>
        <D:propstat>
            <D:prop>
                <Win32CreationTime xmlns="urn:schemas-microsoft-com:"></Win32CreationTime>
                <Win32LastAccessTime xmlns="urn:schemas-microsoft-com:"></Win32LastAccessTime>
                <Win32LastModifiedTime xmlns="urn:schemas-microsoft-com:"></Win32LastModifiedTime>
                <Win32FileAttributes xmlns="urn:schemas-microsoft-com:"></Win32FileAttributes>
            </D:prop>
            <D:status>HTTP/1.1 403 Forbidden</D:status>
        </D:propstat>
    </D:response>
</D:multistatus>"""

        txt = re.sub("\n +", "\n", txt)
        root = mkenod("a")
        root.insert(0, parse_xml(txt))
        prop = root.find(r"./{DAV:}propertyupdate/{DAV:}set/{DAV:}prop")
        assert prop is not None
        assert len(prop)
        for el in prop:
            el.clear()

        res = ET.tostring(prop).decode("utf-8")
        want = """<D:prop xmlns:D="DAV:" xmlns:ns1="urn:schemas-microsoft-com:">
<ns1:Win32CreationTime /><ns1:Win32LastAccessTime /><ns1:Win32LastModifiedTime /><ns1:Win32FileAttributes /></D:prop>
"""
        self.assertEqual(res, want)

    def test5(self):
        txt = r"""<?xml version="1.0" encoding="utf-8" ?>
<D:lockinfo xmlns:D="DAV:">
    <D:lockscope><D:exclusive/></D:lockscope>
    <D:locktype><D:write/></D:locktype>
    <D:owner><D:href>DESKTOP-FRS9AO2\ed</D:href></D:owner>
</D:lockinfo>"""

        ref = r"""<?xml version="1.0" encoding="utf-8"?>
<D:prop xmlns:D="DAV:"><D:lockdiscovery><D:activelock>
    <D:locktype><D:write/></D:locktype>
    <D:lockscope><D:exclusive/></D:lockscope>
    <D:depth>infinity</D:depth>
    <D:owner><D:href>DESKTOP-FRS9AO2\ed</D:href></D:owner>
    <D:timeout>Second-3600</D:timeout>
    <D:locktoken><D:href>1666199679</D:href></D:locktoken>
    <D:lockroot><D:href>/d1/foo.txt</D:href></D:lockroot>
</D:activelock></D:lockdiscovery></D:prop>"""

        txt = re.sub("\n +", "\n", txt)
        lk = parse_xml(txt)
        self.assertEqual(lk.tag, "{DAV:}lockinfo")

        if lk.find(r"./{DAV:}depth") is None:
            lk.append(mktnod("D:depth", "infinity"))

        lk.append(mkenod("D:timeout", mktnod("D:href", "Second-3600")))
        lk.append(mkenod("D:locktoken", mktnod("D:href", "56709")))
        lk.append(mkenod("D:lockroot", mktnod("D:href", "/foo/bar.txt")))

        lk2 = mkenod("D:activelock")
        root = mkenod("D:prop", mkenod("D:lockdiscovery", lk2))
        for a in lk:
            lk2.append(a)

        print(ET.tostring(root).decode("utf-8"))



#!/usr/bin/env python3

import sys
import magic


_ = r"""
reject file uploads by mimetype

dependencies (linux, macos):
    python3 -m pip install --user -U python-magic

dependencies (windows):
    python3 -m pip install --user -U python-magic-bin

example usage as global config:
    --xau c,bin/hooks/reject-mimetype.py

example usage as a volflag (per-volume config):
    -v srv/inc:inc:r:rw,ed:c,xau=c,bin/hooks/reject-mimetype.py
                           ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

    (share filesystem-path srv/inc as volume /inc,
     readable by everyone, read-write for user 'ed',
     running this plugin on all uploads with the params listed below)

parameters explained,
    xau = execute after upload
    c   = check result, reject upload if error
"""


def main():
    ok = ["image/jpeg", "image/png"]

    mt = magic.from_file(sys.argv[1], mime=True)

    print(mt)

    sys.exit(1 if mt not in ok else 0)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import sys
import json
import struct
import base64
import hashlib

try:
    from zlib_ng import zlib_ng as zlib
except:
    import zlib

try:
    from copyparty.util import fsenc
except:

    def fsenc(p):
        return p


"""
calculates various checksums for uploads,
usage: -mtp crc32,md5,sha1,sha256b=ad,bin/mtag/cksum.py
"""


def main():
    config = "crc32 md5 md5b sha1 sha1b sha256 sha256b sha512/240 sha512b/240"
    # b suffix = base64 encoded
    # slash = truncate to n bits

    known = {
        "md5": hashlib.md5,
        "sha1": hashlib.sha1,
        "sha256": hashlib.sha256,
        "sha512": hashlib.sha512,
    }
    config = config.split()
    hashers = {
        k: v()
        for k, v in known.items()
        if k in [x.split("/")[0].rstrip("b") for x in known]
    }
    crc32 = 0 if "crc32" in config else None

    with open(fsenc(sys.argv[1]), "rb", 512 * 1024) as f:
        while True:
            buf = f.read(64 * 1024)
            if not buf:
                break

            for x in hashers.values():
                x.update(buf)

            if crc32 is not None:
                crc32 = zlib.crc32(buf, crc32)

    ret = {}
    for s in config:
        alg = s.split("/")[0]
        b64 = alg.endswith("b")
        alg = alg.rstrip("b")
        if alg in hashers:
            v = hashers[alg].digest()
        elif alg == "crc32":
            v = crc32
            if v < 0:
                v &= 2 ** 32 - 1
            v = struct.pack(">L", v)
        else:
            raise Exception("what is {}".format(s))

        if "/" in s:
            v = v[: int(int(s.split("/")[1]) / 8)]

        if b64:
            v = base64.b64encode(v).decode("ascii").rstrip("=")
        else:
            try:
                v = v.hex()
            except:
                import binascii

                v = binascii.hexlify(v)

        ret[s] = v

    print(json.dumps(ret, indent=4))


if __name__ == "__main__":
    main()



#!/usr/bin/env python

import os
import sys
import vamp
import tempfile
import numpy as np
import subprocess as sp

from copyparty.util import fsenc

"""
dep: vamp
dep: beatroot-vamp
dep: ffmpeg
"""


# save beat timestamps to ".beats/filename.txt"
SAVE = False


def det(tf):
    # fmt: off
    sp.check_call([
        b"ffmpeg",
        b"-nostdin",
        b"-hide_banner",
        b"-v", b"fatal",
        b"-y", b"-i", fsenc(sys.argv[1]),
        b"-map", b"0:a:0",
        b"-ac", b"1",
        b"-ar", b"22050",
        b"-t", b"360",
        b"-f", b"f32le",
        fsenc(tf)
    ])
    # fmt: on

    with open(tf, "rb") as f:
        d = np.fromfile(f, dtype=np.float32)
        try:
            # 98% accuracy on jcore
            c = vamp.collect(d, 22050, "beatroot-vamp:beatroot")
            cl = c["list"]
        except:
            # fallback; 73% accuracy
            plug = "vamp-example-plugins:fixedtempo"
            c = vamp.collect(d, 22050, plug, parameters={"maxdflen": 40})
            print(c["list"][0]["label"].split(" ")[0])
            return

    # throws if detection failed:
    beats = [float(x["timestamp"]) for x in cl]
    bds = [b - a for a, b in zip(beats, beats[1:])]
    bds.sort()
    n0 = int(len(bds) * 0.2)
    n1 = int(len(bds) * 0.75) + 1
    bds = bds[n0:n1]
    bpm = sum(bds)
    bpm = round(60 * (len(bds) / bpm), 2)
    print(f"{bpm:.2f}")

    if SAVE:
        fdir, fname = os.path.split(sys.argv[1])
        bdir = os.path.join(fdir, ".beats")
        try:
            os.mkdir(fsenc(bdir))
        except:
            pass

        fp = os.path.join(bdir, fname) + ".txt"
        with open(fsenc(fp), "wb") as f:
            txt = "\n".join([f"{x:.2f}" for x in beats])
            f.write(txt.encode("utf-8"))


def main():
    with tempfile.NamedTemporaryFile(suffix=".pcm", delete=False) as f:
        f.write(b"h")
        tf = f.name

    try:
        det(tf)
    except:
        pass  # mute
    finally:
        os.unlink(tf)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import os
import threading
import subprocess as sp


_ = r"""
use copyparty as a file downloader by POSTing URLs as
application/x-www-form-urlencoded (for example using the
📟 message-to-server-log in the web-ui)

this hook is a modified copy of wget.py, modified to
make it import-safe so it can be run with the 'I' flag,
which speeds up the startup time of the hook by 140x

example usage as global config:
    --xm aw,I,bin/hooks/wget-i.py

parameters explained,
    xm = execute on message-to-server-log
    aw = only users with write-access can use this
    I = import; do not fork / subprocess

example usage as a volflag (per-volume config):
    -v srv/inc:inc:r:rw,ed:c,xm=aw,I,bin/hooks/wget.py
                           ^^^^^^^^^^^^^^^^^^^^^^^^^^^

    (share filesystem-path srv/inc as volume /inc,
     readable by everyone, read-write for user 'ed',
     running this plugin on all messages with the params explained above)

example usage as a volflag in a copyparty config file:
    [/inc]
      srv/inc
      accs:
        r: *
        rw: ed
      flags:
        xm: aw,I,bin/hooks/wget.py

the volflag examples only kicks in if you send the message
while you're in the /inc folder (or any folder below there)

IMPORTANT NOTE:
    because this hook uses the 'I' flag to run inside copyparty,
    many other flags will not work (f,j,c3,t3600 as seen in the
    original wget.py), and furthermore + more importantly we
    need to be EXCEPTIONALLY CAREFUL to avoid side-effects, so
    the os.chdir has been replaced with cwd=dirpath for example
"""


def do_stuff(inf):
    """
    worker function which is executed in another thread to
    avoid blocking copyparty while the download is running,
    since we cannot use the 'f,t3600' hook-flags with 'I'
    """

    # first things first; grab the logger-function which copyparty is letting us borrow
    log = inf["log"]

    url = inf["txt"]
    if url.startswith("upload-queue-empty;"):
        return

    if "://" not in url:
        url = "https://" + url

    proto = url.split("://")[0].lower()
    if proto not in ("http", "https", "ftp", "ftps"):
        raise Exception("bad proto {}".format(proto))

    dirpath = inf["ap"]

    name = url.split("?")[0].split("/")[-1]
    msg = "-- DOWNLOADING " + name
    log(msg)
    tfn = os.path.join(dirpath, msg)
    open(tfn, "wb").close()

    cmd = ["wget", "--trust-server-names", "-nv", "--", url]

    try:
        # two things to note here:
        # - cannot use the `c3` hook-flag with `I` so mute output with stdout=sp.DEVNULL instead;
        # - MUST NOT use os.chdir with 'I' so use cwd=dirpath instead 
        sp.check_call(cmd, cwd=dirpath, stdout=sp.DEVNULL)
    except:
        t = "-- FAILED TO DOWNLOAD " + name
        log(t, 3)  # 3=yellow=warning
        open(os.path.join(dirpath, t), "wb").close()
        raise  # have copyparty scream about the details in the log

    os.unlink(tfn)


def main(inf):
    threading.Thread(target=do_stuff, args=(inf,), daemon=True).start()



#!/usr/bin/env python3
from __future__ import print_function, unicode_literals

"""partyfuse2: remote copyparty as a local filesystem"""
__author__ = "ed <copyparty@ocv.me>"
__copyright__ = 2020
__license__ = "MIT"
__url__ = "https://github.com/9001/copyparty/"

import re
import os
import sys
import time
import json
import stat
import errno
import struct
import codecs
import platform
import threading
import http.client  # py2: httplib
import urllib.parse
from datetime import datetime
from urllib.parse import quote_from_bytes as quote
from urllib.parse import unquote_to_bytes as unquote

try:
    import fuse
    from fuse import Fuse

    fuse.fuse_python_api = (0, 2)
    if not hasattr(fuse, "__version__"):
        raise Exception("your fuse-python is way old")
except:
    if WINDOWS:
        libfuse = "install https://github.com/billziss-gh/winfsp/releases/latest"
    elif MACOS:
        libfuse = "install https://osxfuse.github.io/"
    else:
        libfuse = "apt install libfuse\n    modprobe fuse"

    m = """\033[33m
  could not import fuse; these may help:
    {} -m pip install --user fuse-python
    {}
\033[0m"""
    print(m.format(sys.executable, libfuse))
    raise


"""
mount a copyparty server (local or remote) as a filesystem

usage:
  python ./partyfuse2.py -f -o allow_other,auto_unmount,nonempty,pw=wark,url=http://192.168.1.69:3923 /mnt/nas

dependencies:
  sudo apk add fuse-dev python3-dev
  python3 -m pip install --user fuse-python

fork of partyfuse.py based on fuse-python which
  appears to be more compliant than fusepy? since this works with samba
    (probably just my garbage code tbh)
"""


WINDOWS = sys.platform == "win32"
MACOS = platform.system() == "Darwin"


def threadless_log(msg):
    print(msg + "\n", end="")


def boring_log(msg):
    msg = "\033[36m{:012x}\033[0m {}\n".format(threading.current_thread().ident, msg)
    print(msg[4:], end="")


def rice_tid():
    tid = threading.current_thread().ident
    c = struct.unpack(b"B" * 5, struct.pack(b">Q", tid)[-5:])
    return "".join("\033[1;37;48;5;{}m{:02x}".format(x, x) for x in c) + "\033[0m"


def fancy_log(msg):
    print("{} {}\n".format(rice_tid(), msg), end="")


def null_log(msg):
    pass


info = fancy_log
log = fancy_log
dbg = fancy_log
log = null_log
dbg = null_log


def get_tid():
    return threading.current_thread().ident


def html_dec(txt):
    return (
        txt.replace("&lt;", "<")
        .replace("&gt;", ">")
        .replace("&quot;", '"')
        .replace("&amp;", "&")
    )


def register_wtf8():
    def wtf8_enc(text):
        return str(text).encode("utf-8", "surrogateescape"), len(text)

    def wtf8_dec(binary):
        return bytes(binary).decode("utf-8", "surrogateescape"), len(binary)

    def wtf8_search(encoding_name):
        return codecs.CodecInfo(wtf8_enc, wtf8_dec, name="wtf-8")

    codecs.register(wtf8_search)


bad_good = {}
good_bad = {}


def enwin(txt):
    return "".join([bad_good.get(x, x) for x in txt])

    for bad, good in bad_good.items():
        txt = txt.replace(bad, good)

    return txt


def dewin(txt):
    return "".join([good_bad.get(x, x) for x in txt])

    for bad, good in bad_good.items():
        txt = txt.replace(good, bad)

    return txt


class CacheNode(object):
    def __init__(self, tag, data):
        self.tag = tag
        self.data = data
        self.ts = time.time()


class Stat(fuse.Stat):
    def __init__(self):
        self.st_mode = 0
        self.st_ino = 0
        self.st_dev = 0
        self.st_nlink = 1
        self.st_uid = 1000
        self.st_gid = 1000
        self.st_size = 0
        self.st_atime = 0
        self.st_mtime = 0
        self.st_ctime = 0


class Gateway(object):
    def __init__(self, base_url, pw):
        self.base_url = base_url
        self.pw = pw

        ui = urllib.parse.urlparse(base_url)
        self.web_root = ui.path.strip("/")
        try:
            self.web_host, self.web_port = ui.netloc.split(":")
            self.web_port = int(self.web_port)
        except:
            self.web_host = ui.netloc
            if ui.scheme == "http":
                self.web_port = 80
            elif ui.scheme == "https":
                raise Exception("todo")
            else:
                raise Exception("bad url?")

        self.conns = {}

    def quotep(self, path):
        path = path.encode("wtf-8")
        return quote(path, safe="/")

    def getconn(self, tid=None):
        tid = tid or get_tid()
        try:
            return self.conns[tid]
        except:
            info("new conn [{}] [{}]".format(self.web_host, self.web_port))

            conn = http.client.HTTPConnection(self.web_host, self.web_port, timeout=260)

            self.conns[tid] = conn
            return conn

    def closeconn(self, tid=None):
        tid = tid or get_tid()
        try:
            self.conns[tid].close()
            del self.conns[tid]
        except:
            pass

    def sendreq(self, *args, **ka):
        tid = get_tid()
        if self.pw:
            ck = "cppwd=" + self.pw
            try:
                ka["headers"]["Cookie"] = ck
            except:
                ka["headers"] = {"Cookie": ck}
        try:
            c = self.getconn(tid)
            c.request(*list(args), **ka)
            return c.getresponse()
        except:
            self.closeconn(tid)
            c = self.getconn(tid)
            c.request(*list(args), **ka)
            return c.getresponse()

    def listdir(self, path):
        if bad_good:
            path = dewin(path)

        web_path = self.quotep("/" + "/".join([self.web_root, path])) + "?dots&ls"
        r = self.sendreq("GET", web_path)
        if r.status != 200:
            self.closeconn()
            raise Exception(
                "http error {} reading dir {} in {}".format(
                    r.status, web_path, rice_tid()
                )
            )

        return self.parse_jls(r)

    def download_file_range(self, path, ofs1, ofs2):
        if bad_good:
            path = dewin(path)

        web_path = self.quotep("/" + "/".join([self.web_root, path])) + "?raw"
        hdr_range = "bytes={}-{}".format(ofs1, ofs2 - 1)
        log("downloading {}".format(hdr_range))

        r = self.sendreq("GET", web_path, headers={"Range": hdr_range})
        if r.status != http.client.PARTIAL_CONTENT:
            self.closeconn()
            raise Exception(
                "http error {} reading file {} range {} in {}".format(
                    r.status, web_path, hdr_range, rice_tid()
                )
            )

        return r.read()

    def parse_jls(self, datasrc):
        rsp = b""
        while True:
            buf = datasrc.read(1024 * 32)
            if not buf:
                break

            rsp += buf

        rsp = json.loads(rsp.decode("utf-8"))
        ret = []
        for statfun, nodes in [
            [self.stat_dir, rsp["dirs"]],
            [self.stat_file, rsp["files"]],
        ]:
            for n in nodes:
                fname = unquote(n["href"].split("?")[0]).rstrip(b"/").decode("wtf-8")
                if bad_good:
                    fname = enwin(fname)

                ret.append([fname, statfun(n["ts"], n["sz"]), 0])

        return ret

    def stat_dir(self, ts, sz=4096):
        ret = Stat()
        ret.st_mode = stat.S_IFDIR | 0o555
        ret.st_nlink = 2
        ret.st_size = sz
        ret.st_atime = ts
        ret.st_mtime = ts
        ret.st_ctime = ts
        return ret

    def stat_file(self, ts, sz):
        ret = Stat()
        ret.st_mode = stat.S_IFREG | 0o444
        ret.st_size = sz
        ret.st_atime = ts
        ret.st_mtime = ts
        ret.st_ctime = ts
        return ret


class CPPF(Fuse):
    def __init__(self, *args, **kwargs):
        Fuse.__init__(self, *args, **kwargs)

        self.url = None
        self.pw = None

        self.dircache = []
        self.dircache_mtx = threading.Lock()

        self.filecache = []
        self.filecache_mtx = threading.Lock()

    def init2(self):
        # TODO figure out how python-fuse wanted this to go
        self.gw = Gateway(self.url, self.pw)  # .decode('utf-8'))
        info("up")

    def clean_dircache(self):
        """not threadsafe"""
        now = time.time()
        cutoff = 0
        for cn in self.dircache:
            if now - cn.ts > 1:
                cutoff += 1
            else:
                break

        if cutoff > 0:
            self.dircache = self.dircache[cutoff:]

    def get_cached_dir(self, dirpath):
        # with self.dircache_mtx:
        if True:
            self.clean_dircache()
            for cn in self.dircache:
                if cn.tag == dirpath:
                    return cn

        return None

    """
            ,-------------------------------,  g1>=c1, g2<=c2
            |cache1                   cache2|  buf[g1-c1:(g1-c1)+(g2-g1)]
            `-------------------------------'
                    ,---------------,
                    |get1       get2|
                    `---------------'
    __________________________________________________________________________

            ,-------------------------------,  g2<=c2, (g2>=c1)
            |cache1                   cache2|  cdr=buf[:g2-c1]
            `-------------------------------'  dl car; g1-512K:c1
    ,---------------,
    |get1       get2|
    `---------------'
    __________________________________________________________________________

            ,-------------------------------,  g1>=c1, (g1<=c2)
            |cache1                   cache2|  car=buf[c2-g1:]
            `-------------------------------'  dl cdr; c2:c2+1M
                                    ,---------------,
                                    |get1       get2|
                                    `---------------'
    """

    def get_cached_file(self, path, get1, get2, file_sz):
        car = None
        cdr = None
        ncn = -1
        # with self.filecache_mtx:
        if True:
            dbg("cache request from {} to {}, size {}".format(get1, get2, file_sz))
            for cn in self.filecache:
                ncn += 1

                cache_path, cache1 = cn.tag
                if cache_path != path:
                    continue

                cache2 = cache1 + len(cn.data)
                if get2 <= cache1 or get1 >= cache2:
                    continue

                if get1 >= cache1 and get2 <= cache2:
                    # keep cache entry alive by moving it to the end
                    self.filecache = (
                        self.filecache[:ncn] + self.filecache[ncn + 1 :] + [cn]
                    )
                    buf_ofs = get1 - cache1
                    buf_end = buf_ofs + (get2 - get1)
                    dbg(
                        "found all ({}, {} to {}, len {}) [{}:{}] = {}".format(
                            ncn,
                            cache1,
                            cache2,
                            len(cn.data),
                            buf_ofs,
                            buf_end,
                            buf_end - buf_ofs,
                        )
                    )
                    return cn.data[buf_ofs:buf_end]

                if get2 < cache2:
                    x = cn.data[: get2 - cache1]
                    if not cdr or len(cdr) < len(x):
                        dbg(
                            "found car ({}, {} to {}, len {}) [:{}-{}] = [:{}] = {}".format(
                                ncn,
                                cache1,
                                cache2,
                                len(cn.data),
                                get2,
                                cache1,
                                get2 - cache1,
                                len(x),
                            )
                        )
                        cdr = x

                    continue

                if get1 > cache1:
                    x = cn.data[-(cache2 - get1) :]
                    if not car or len(car) < len(x):
                        dbg(
                            "found cdr ({}, {} to {}, len {}) [-({}-{}):] = [-{}:] = {}".format(
                                ncn,
                                cache1,
                                cache2,
                                len(cn.data),
                                cache2,
                                get1,
                                cache2 - get1,
                                len(x),
                            )
                        )
                        car = x

                    continue

                raise Exception("what")

        if car and cdr:
            dbg("<cache> have both")

            ret = car + cdr
            if len(ret) == get2 - get1:
                return ret

            raise Exception("{} + {} != {} - {}".format(len(car), len(cdr), get2, get1))

        elif cdr:
            h_end = get1 + (get2 - get1) - len(cdr)
            h_ofs = h_end - 512 * 1024

            if h_ofs < 0:
                h_ofs = 0

            buf_ofs = (get2 - get1) - len(cdr)

            dbg(
                "<cache> cdr {}, car {}-{}={} [-{}:]".format(
                    len(cdr), h_ofs, h_end, h_end - h_ofs, buf_ofs
                )
            )

            buf = self.gw.download_file_range(path, h_ofs, h_end)
            ret = buf[-buf_ofs:] + cdr

        elif car:
            h_ofs = get1 + len(car)
            h_end = h_ofs + 1024 * 1024

            if h_end > file_sz:
                h_end = file_sz

            buf_ofs = (get2 - get1) - len(car)

            dbg(
                "<cache> car {}, cdr {}-{}={} [:{}]".format(
                    len(car), h_ofs, h_end, h_end - h_ofs, buf_ofs
                )
            )

            buf = self.gw.download_file_range(path, h_ofs, h_end)
            ret = car + buf[:buf_ofs]

        else:
            h_ofs = get1 - 256 * 1024
            h_end = get2 + 1024 * 1024

            if h_ofs < 0:
                h_ofs = 0

            if h_end > file_sz:
                h_end = file_sz

            buf_ofs = get1 - h_ofs
            buf_end = buf_ofs + get2 - get1

            dbg(
                "<cache> {}-{}={} [{}:{}]".format(
                    h_ofs, h_end, h_end - h_ofs, buf_ofs, buf_end
                )
            )

            buf = self.gw.download_file_range(path, h_ofs, h_end)
            ret = buf[buf_ofs:buf_end]

        cn = CacheNode([path, h_ofs], buf)
        # with self.filecache_mtx:
        if True:
            if len(self.filecache) > 6:
                self.filecache = self.filecache[1:] + [cn]
            else:
                self.filecache.append(cn)

        return ret

    def _readdir(self, path):
        path = path.strip("/")
        log("readdir {}".format(path))

        ret = self.gw.listdir(path)

        # with self.dircache_mtx:
        if True:
            cn = CacheNode(path, ret)
            self.dircache.append(cn)
            self.clean_dircache()

        return ret

    def readdir(self, path, offset):
        for e in self._readdir(path)[offset:]:
            # log("yield [{}]".format(e[0]))
            yield fuse.Direntry(e[0])

    def open(self, path, flags):
        if (flags & (os.O_RDONLY | os.O_WRONLY | os.O_RDWR)) != os.O_RDONLY:
            return -errno.EACCES

        st = self.getattr(path)
        try:
            if st.st_nlink > 0:
                return st
        except:
            return st  # -int(os.errcode)

    def read(self, path, length, offset, fh=None, *args):
        if args:
            log("unexpected args [" + "] [".join(repr(x) for x in args) + "]")
            raise Exception()

        path = path.strip("/")

        ofs2 = offset + length
        log("read {} @ {} len {} end {}".format(path, offset, length, ofs2))

        st = self.getattr(path)
        try:
            file_sz = st.st_size
        except:
            return st  # -int(os.errcode)

        if ofs2 > file_sz:
            ofs2 = file_sz
            log("truncate to len {} end {}".format(ofs2 - offset, ofs2))

        if file_sz == 0 or offset >= ofs2:
            return b""

        # toggle cache here i suppose
        # return self.get_cached_file(path, offset, ofs2, file_sz)
        return self.gw.download_file_range(path, offset, ofs2)

    def getattr(self, path):
        log("getattr [{}]".format(path))
        if WINDOWS:
            path = enwin(path)  # windows occasionally decodes f0xx to xx

        path = path.strip("/")
        try:
            dirpath, fname = path.rsplit("/", 1)
        except:
            dirpath = ""
            fname = path

        if not path:
            ret = self.gw.stat_dir(time.time())
            dbg("=root")
            return ret

        cn = self.get_cached_dir(dirpath)
        if cn:
            log("cache ok")
            dents = cn.data
        else:
            log("cache miss")
            dents = self._readdir(dirpath)

        for cache_name, cache_stat, _ in dents:
            if cache_name == fname:
                dbg("=file")
                return cache_stat

        log("=404")
        return -errno.ENOENT


def main():
    time.strptime("19970815", "%Y%m%d")  # python#7980
    register_wtf8()
    if WINDOWS:
        os.system("rem")

        for ch in '<>:"\\|?*':
            # microsoft maps illegal characters to f0xx
            # (e000 to f8ff is basic-plane private-use)
            bad_good[ch] = chr(ord(ch) + 0xF000)

        for n in range(0, 0x100):
            # map surrogateescape to another private-use area
            bad_good[chr(n + 0xDC00)] = chr(n + 0xF100)

        for k, v in bad_good.items():
            good_bad[v] = k

    server = CPPF()
    server.parser.add_option(mountopt="url", metavar="BASE_URL", default=None)
    server.parser.add_option(mountopt="pw", metavar="PASSWORD", default=None)
    server.parse(values=server, errex=1)
    if not server.url or not str(server.url).startswith("http"):
        print("\nerror:")
        print("  need argument: -o url=<...>")
        print("  need argument: mount-path")
        print("example:")
        print(
            "  ./partyfuse2.py -f -o allow_other,auto_unmount,nonempty,pw=wark,url=http://192.168.1.69:3923 /mnt/nas"
        )
        sys.exit(1)

    server.init2()
    threading.Thread(target=server.main, daemon=True).start()
    while True:
        time.sleep(9001)


if __name__ == "__main__":
    main()



#!/usr/bin/env python

import re
import os
import sys
import gzip
import json
import base64
import string
import urllib.request
from datetime import datetime

"""
youtube initial player response

it's probably best to use this through a config file; see res/yt-ipr.conf

but if you want to use plain arguments instead then:
  -v srv/ytm:ytm:w:rw,ed
       :c,e2ts,e2dsa
       :c,sz=16k-1m:c,maxn=10,300:c,rotf=%Y-%m/%d-%H
       :c,mtp=yt-id,yt-title,yt-author,yt-channel,yt-views,yt-private,yt-manifest,yt-expires=bin/mtag/yt-ipr.py
       :c,mte=yt-id,yt-title,yt-author,yt-channel,yt-views,yt-private,yt-manifest,yt-expires

see res/yt-ipr.user.js for the example userscript to go with this
"""


def main():
    try:
        with gzip.open(sys.argv[1], "rt", encoding="utf-8", errors="replace") as f:
            txt = f.read()
    except:
        with open(sys.argv[1], "r", encoding="utf-8", errors="replace") as f:
            txt = f.read()

    txt = "{" + txt.split("{", 1)[1]

    try:
        pd = json.loads(txt)
    except json.decoder.JSONDecodeError as ex:
        pd = json.loads(txt[: ex.pos])

    # print(json.dumps(pd, indent=2))

    if "videoDetails" in pd:
        parse_youtube(pd)
    else:
        parse_freg(pd)


def get_expiration(url):
    et = re.search(r"[?&]expire=([0-9]+)", url).group(1)
    et = datetime.utcfromtimestamp(int(et))
    return et.strftime("%Y-%m-%d, %H:%M")


def parse_youtube(pd):
    vd = pd["videoDetails"]
    sd = pd["streamingData"]

    et = sd["adaptiveFormats"][0]["url"]
    et = get_expiration(et)

    mf = []
    if "dashManifestUrl" in sd:
        mf.append("dash")
    if "hlsManifestUrl" in sd:
        mf.append("hls")

    r = {
        "yt-id": vd["videoId"],
        "yt-title": vd["title"],
        "yt-author": vd["author"],
        "yt-channel": vd["channelId"],
        "yt-views": vd["viewCount"],
        "yt-private": vd["isPrivate"],
        # "yt-expires": sd["expiresInSeconds"],
        "yt-manifest": ",".join(mf),
        "yt-expires": et,
    }
    print(json.dumps(r))

    freg_conv(pd)


def parse_freg(pd):
    md = pd["metadata"]
    r = {
        "yt-id": md["id"],
        "yt-title": md["title"],
        "yt-author": md["channelName"],
        "yt-channel": md["channelURL"].strip("/").split("/")[-1],
        "yt-expires": get_expiration(list(pd["video"].values())[0]),
    }
    print(json.dumps(r))


def freg_conv(pd):
    # based on getURLs.js v1.5 (2021-08-07)
    # fmt: off
    priority = {
        "video": [
            337, 315, 266, 138,  # 2160p60
            313, 336,  # 2160p
            308,  # 1440p60
            271, 264,  # 1440p
            335, 303, 299,  # 1080p60
            248, 169, 137,  # 1080p
            334, 302, 298,  # 720p60
            247, 136  # 720p
        ],
        "audio": [
            251, 141, 171, 140, 250, 249, 139
        ]
    }

    vid_id = pd["videoDetails"]["videoId"]
    chan_id = pd["videoDetails"]["channelId"]

    try:
        thumb_url = pd["microformat"]["playerMicroformatRenderer"]["thumbnail"]["thumbnails"][0]["url"]
        start_ts = pd["microformat"]["playerMicroformatRenderer"]["liveBroadcastDetails"]["startTimestamp"]
    except:
        thumb_url = f"https://img.youtube.com/vi/{vid_id}/maxresdefault.jpg"
        start_ts = ""

    # fmt: on

    metadata = {
        "title": pd["videoDetails"]["title"],
        "id": vid_id,
        "channelName": pd["videoDetails"]["author"],
        "channelURL": "https://www.youtube.com/channel/" + chan_id,
        "description": pd["videoDetails"]["shortDescription"],
        "thumbnailUrl": thumb_url,
        "startTimestamp": start_ts,
    }

    if [x for x in vid_id if x not in string.ascii_letters + string.digits + "_-"]:
        print(f"malicious json", file=sys.stderr)
        return

    basepath = os.path.dirname(sys.argv[1])

    thumb_fn = f"{basepath}/{vid_id}.jpg"
    tmp_fn = f"{thumb_fn}.{os.getpid()}"
    if not os.path.exists(thumb_fn) and (
        thumb_url.startswith("https://img.youtube.com/vi/")
        or thumb_url.startswith("https://i.ytimg.com/vi/")
    ):
        try:
            with urllib.request.urlopen(thumb_url) as fi:
                with open(tmp_fn, "wb") as fo:
                    fo.write(fi.read())

            os.rename(tmp_fn, thumb_fn)
        except:
            if os.path.exists(tmp_fn):
                os.unlink(tmp_fn)

    try:
        with open(thumb_fn, "rb") as f:
            thumb = base64.b64encode(f.read()).decode("ascii")
    except:
        thumb = "/9j/4AAQSkZJRgABAQEASABIAAD/2wBDAAMCAgICAgMCAgIDAwMDBAYEBAQEBAgGBgUGCQgKCgkICQkKDA8MCgsOCwkJDRENDg8QEBEQCgwSExIQEw8QEBD/yQALCAABAAEBAREA/8wABgAQEAX/2gAIAQEAAD8A0s8g/9k="

    metadata["thumbnail"] = "data:image/jpeg;base64," + thumb

    ret = {
        "metadata": metadata,
        "version": "1.5",
        "createTime": datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%SZ"),
    }

    for stream, itags in priority.items():
        for itag in itags:
            url = None
            for afmt in pd["streamingData"]["adaptiveFormats"]:
                if itag == afmt["itag"]:
                    url = afmt["url"]
                    break

            if url:
                ret[stream] = {itag: url}
                break

    fn = f"{basepath}/{vid_id}.urls.json"
    with open(fn, "w", encoding="utf-8", errors="replace") as f:
        f.write(json.dumps(ret, indent=4))


if __name__ == "__main__":
    try:
        main()
    except:
        # raise
        pass



#!/usr/bin/env python3

# Update the Nix package pin
#
# Usage: ./update.sh

import base64
import json
import hashlib
import sys
from pathlib import Path

OUTPUT_FILE = Path("pin.json")
TARGET_ASSET = lambda version: f"partftpy-{version}.tar.gz"
HASH_TYPE = "sha256"
LATEST_RELEASE_URL = "https://api.github.com/repos/9001/partftpy/releases/latest"


def get_formatted_hash(binary):
    hasher = hashlib.new("sha256")
    hasher.update(binary)
    asset_hash = hasher.digest()
    encoded_hash = base64.b64encode(asset_hash).decode("ascii")
    return f"{HASH_TYPE}-{encoded_hash}"


def remote_release_pin():
    import requests

    response = requests.get(LATEST_RELEASE_URL).json()
    version = response["tag_name"].lstrip("v")
    asset_info = [a for a in response["assets"] if a["name"] == TARGET_ASSET(version)][0]
    download_url = asset_info["browser_download_url"]
    asset = requests.get(download_url)
    formatted_hash = get_formatted_hash(asset.content)

    result = {"url": download_url, "version": version, "hash": formatted_hash}
    return result


def main():
    result = remote_release_pin()

    print(result)
    json_result = json.dumps(result, indent=4)
    OUTPUT_FILE.write_text(json_result)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

"""
fetch latest msg from guestbook and return as tag

example copyparty config to use this:
  --urlform save,get -vsrv/hello:hello:w:c,e2ts,mtp=guestbook=t10,ad,p,bin/mtag/guestbook-read.py:mte=+guestbook

explained:
  for realpath srv/hello (served at /hello), write-only for everyone,
  enable file analysis on upload (e2ts),
  use mtp plugin "bin/mtag/guestbook-read.py" to provide metadata tag "guestbook",
  do this on all uploads regardless of extension,
  t10 = 10 seconds timeout for each dwonload,
  ad = parse file regardless if FFmpeg thinks it is audio or not
  p = request upload info as json on stdin (need ip)
  mte=+guestbook enabled indexing of that tag for this volume

PS: this requires e2ts to be functional,
  meaning you need to do at least one of these:
   * apt install ffmpeg
   * pip3 install mutagen
"""


import json
import os
import sqlite3
import sys


# set 0 to allow infinite msgs from one IP,
# other values delete older messages to make space,
# so 1 only keeps latest msg
NUM_MSGS_TO_KEEP = 1


def main():
    fp = os.path.abspath(sys.argv[1])
    fdir = os.path.dirname(fp)

    zb = sys.stdin.buffer.read()
    zs = zb.decode("utf-8", "replace")
    md = json.loads(zs)

    ip = md["up_ip"]

    # can put the database inside `fdir` if you'd like,
    # by default it saves to PWD:
    # os.chdir(fdir)

    db = sqlite3.connect("guestbook.db3")
    with db:
        t = "select msg from gb where ip = ? order by ts desc"
        r = db.execute(t, (ip,)).fetchone()
        if r:
            print(r[0])


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import os
import sys
import stat
import time
import signal
import traceback
import threading
from queue import Queue


"""speedtest-fs: filesystem performance estimate"""
__author__ = "ed <copyparty@ocv.me>"
__copyright__ = 2020
__license__ = "MIT"
__url__ = "https://github.com/9001/copyparty/"


def get_spd(nbyte, nfiles, nsec):
    if not nsec:
        return "0.000 MB   0 files   0.000 sec   0.000 MB/s   0.000 f/s"

    mb = nbyte / (1024 * 1024.0)
    spd = mb / nsec
    nspd = nfiles / nsec

    return f"{mb:.3f} MB   {nfiles} files   {nsec:.3f} sec   {spd:.3f} MB/s   {nspd:.3f} f/s"


class Inf(object):
    def __init__(self, t0):
        self.msgs = []
        self.errors = []
        self.reports = []
        self.mtx_msgs = threading.Lock()
        self.mtx_reports = threading.Lock()

        self.n_byte = 0
        self.n_file = 0
        self.n_sec = 0
        self.n_done = 0
        self.t0 = t0

        thr = threading.Thread(target=self.print_msgs)
        thr.daemon = True
        thr.start()

    def msg(self, fn, n_read):
        with self.mtx_msgs:
            self.msgs.append(f"{fn} {n_read}")

    def err(self, fn):
        with self.mtx_reports:
            self.errors.append(f"{fn}\n{traceback.format_exc()}")

    def print_msgs(self):
        while True:
            time.sleep(0.02)
            with self.mtx_msgs:
                msgs = self.msgs
                self.msgs = []

            if not msgs:
                continue

            msgs = msgs[-64:]
            spd = get_spd(self.n_byte, len(self.reports), self.n_sec)
            msgs = [f"{spd}   {x}" for x in msgs]
            print("\n".join(msgs))

    def report(self, fn, n_byte, n_sec):
        with self.mtx_reports:
            self.reports.append([n_byte, n_sec, fn])
            self.n_byte += n_byte
            self.n_sec += n_sec

    def done(self):
        with self.mtx_reports:
            self.n_done += 1


def get_files(dir_path):
    for fn in os.listdir(dir_path):
        fn = os.path.join(dir_path, fn)
        st = os.stat(fn).st_mode

        if stat.S_ISDIR(st):
            yield from get_files(fn)

        if stat.S_ISREG(st):
            yield fn


def worker(q, inf, read_sz):
    while True:
        fn = q.get()
        if not fn:
            break

        n_read = 0
        try:
            t0 = time.time()
            with open(fn, "rb") as f:
                while True:
                    buf = f.read(read_sz)
                    if not buf:
                        break

                    n_read += len(buf)
                    inf.msg(fn, n_read)

            inf.report(fn, n_read, time.time() - t0)
        except:
            inf.err(fn)

    inf.done()


def sighandler(signo, frame):
    os._exit(0)


def main():
    signal.signal(signal.SIGINT, sighandler)

    root = "."
    if len(sys.argv) > 1:
        root = sys.argv[1]

    t0 = time.time()
    q = Queue(256)
    inf = Inf(t0)

    num_threads = 8
    read_sz = 32 * 1024
    targs = (q, inf, read_sz)
    for _ in range(num_threads):
        thr = threading.Thread(target=worker, args=targs)
        thr.daemon = True
        thr.start()

    for fn in get_files(root):
        q.put(fn)

    for _ in range(num_threads):
        q.put(None)

    while inf.n_done < num_threads:
        time.sleep(0.1)

    t2 = time.time()
    print("\n")

    log = inf.reports
    log.sort()
    for nbyte, nsec, fn in log[-64:]:
        spd = get_spd(nbyte, len(log), nsec)
        print(f"{spd}   {fn}")

    print()
    print("\n".join(inf.errors))

    print(get_spd(inf.n_byte, len(log), t2 - t0))


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import itertools
import re
import sys
import time

from copyparty.authsrv import AuthSrv
from copyparty.httpcli import HttpCli

from . import util as tu
from .util import Cfg

atlas = ["%", "25", "2e", "2f", ".", "/"]


def nolog(*a, **ka):
    pass


def hdr(query):
    h = "GET /{} HTTP/1.1\r\nCookie: cppwd=o\r\nConnection: close\r\n\r\n"
    return h.format(query).encode("utf-8")


def curl(args, asrv, url, binary=False):
    conn = tu.VHttpConn(args, asrv, nolog, hdr(url))
    HttpCli(conn).run()
    if binary:
        h, b = conn.s._reply.split(b"\r\n\r\n", 1)
        return [h.decode("utf-8"), b]

    return conn.s._reply.decode("utf-8").split("\r\n\r\n", 1)


def genlen(ubase, ntot, nth, wlen):
    args = Cfg(v=["s2::r"], a=["o:o", "x:x"])
    asrv = AuthSrv(args, print)
    # h, ret = curl(args, asrv, "hey")

    n = 0
    t0 = time.time()
    print("genlen %s nth %s" % (wlen, nth))
    ptn = re.compile(r"2.2.2.2|\.\.\.|///|%%%|\.2|/2./|%\.|/%/")
    for path in itertools.product(atlas, repeat=wlen):
        if "%" not in path:
            continue
        path = "".join(path)
        if ptn.search(path):
            continue
        n += 1
        if n % ntot != nth:
            continue
        url = ubase + path + "fa"
        if n % 500 == nth:
            spd = n / (time.time() - t0)
            print(wlen, n, int(spd), url)

        hdr, r = curl(args, asrv, url)
        if "fgsfds" in r:
            with open("hit-%s.txt" % (time.time()), "w", encoding="utf-8") as f:
                f.write(url)
            raise Exception("HIT! {}".format(url))


def main():
    ubase = sys.argv[1]
    ntot = int(sys.argv[2])
    nth = int(sys.argv[3])
    for wlen in range(20):
        genlen(ubase, ntot, nth, wlen)


if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        pass


"""
nice pypy3 -m tests.ptrav "" 2 0
nice pypy3 -m tests.ptrav "" 2 1
nice pypy3 -m tests.ptrav .cpr 2 0
nice pypy3 -m tests.ptrav .cpr 2 1
(13x faster than /scripts/test/ptrav.py)
"""



# coding: utf-8
from __future__ import print_function, unicode_literals

import argparse

from queue import Queue

from .__init__ import TYPE_CHECKING
from .authsrv import AuthSrv
from .util import HMaccas, Pebkac

if True:  # pylint: disable=using-constant-test
    from typing import Any, Optional, Union

    from .util import RootLogger

if TYPE_CHECKING:
    from .httpsrv import HttpSrv
    from .svchub import SvcHub


class ExceptionalQueue(Queue, object):
    def get(self, block: bool = True, timeout: Optional[float] = None) -> Any:
        rv = super(ExceptionalQueue, self).get(block, timeout)

        if isinstance(rv, list):
            if rv[0] == "exception":
                if rv[1] == "pebkac":
                    raise Pebkac(*rv[2:])
                else:
                    raise rv[2]

        return rv


class NotExQueue(object):
    """
    BrokerThr uses this instead of ExceptionalQueue; 7x faster
    """

    def __init__(self, rv: Any) -> None:
        self.rv = rv

    def get(self) -> Any:
        return self.rv


class BrokerCli(object):
    """
    helps mypy understand httpsrv.broker but still fails a few levels deeper,
    for example resolving httpconn.* in httpcli -- see lines tagged #mypy404
    """

    hub: "SvcHub"
    log: "RootLogger"
    args: argparse.Namespace
    asrv: AuthSrv
    httpsrv: "HttpSrv"
    iphash: HMaccas

    def __init__(self) -> None:
        pass

    def ask(self, dest: str, *args: Any) -> Union[ExceptionalQueue, NotExQueue]:
        return ExceptionalQueue(1)

    def say(self, dest: str, *args: Any) -> None:
        pass


def try_exec(want_retval: Union[bool, int], func: Any, *args: list[Any]) -> Any:
    try:
        return func(*args)

    except Pebkac as ex:
        if not want_retval:
            raise

        return ["exception", "pebkac", ex.code, str(ex)]

    except Exception as ex:
        if not want_retval:
            raise

        return ["exception", "stack", ex]



#!/usr/bin/env python3
# coding: utf-8
from __future__ import print_function, unicode_literals

import os
import shutil
import tempfile
import unittest
from itertools import product

from copyparty.authsrv import AuthSrv
from copyparty.httpcli import HttpCli
from tests import util as tu
from tests.util import Cfg


class TestDedup(tu.TC):
    def setUp(self):
        self.td = tu.get_ramdisk()

    def tearDown(self):
        if self.conn:
            self.conn.shutdown()
        os.chdir(tempfile.gettempdir())
        shutil.rmtree(self.td)

    def reset(self):
        td = os.path.join(self.td, "vfs")
        if os.path.exists(td):
            shutil.rmtree(td)
        os.mkdir(td)
        os.chdir(td)
        for a in "abc":
            os.mkdir(a)
            for b in "fg":
                d = "%s/%s%s" % (a, a, b)
                os.mkdir(d)
                for fn in "x":
                    fp = "%s/%s%s%s" % (d, a, b, fn)
                    with open(fp, "wb") as f:
                        f.write(fp.encode("utf-8"))
        return td

    def cinit(self):
        if self.conn:
            self.fstab = self.conn.hsrv.hub.up2k.fstab
            self.conn.hsrv.hub.up2k.shutdown()
        self.asrv = AuthSrv(self.args, self.log)
        self.conn = tu.VHttpConn(self.args, self.asrv, self.log, b"", True)
        if self.fstab:
            self.conn.hsrv.hub.up2k.fstab = self.fstab

    def test(self):
        tc_dedup = ["sym", "no"]
        vols = [".::A", "a/af:a/af:r", "b:a/b:r"]
        tcs = [
            "/a?copy=/c/a /a/af/afx /a/ag/agx /a/b/bf/bfx /a/b/bg/bgx /b/bf/bfx /b/bg/bgx /c/a/af/afx /c/a/ag/agx /c/a/b/bf/bfx /c/a/b/bg/bgx /c/cf/cfx /c/cg/cgx",
            "/b?copy=/d /a/af/afx /a/ag/agx /a/b/bf/bfx /a/b/bg/bgx /b/bf/bfx /b/bg/bgx /c/cf/cfx /c/cg/cgx /d/bf/bfx /d/bg/bgx",
            "/b/bf?copy=/d /a/af/afx /a/ag/agx /a/b/bf/bfx /a/b/bg/bgx /b/bf/bfx /b/bg/bgx /c/cf/cfx /c/cg/cgx /d/bfx",
            "/a/af?copy=/d /a/af/afx /a/ag/agx /a/b/bf/bfx /a/b/bg/bgx /b/bf/bfx /b/bg/bgx /c/cf/cfx /c/cg/cgx /d/afx",
            "/a/af?copy=/ /a/af/afx /a/ag/agx /a/b/bf/bfx /a/b/bg/bgx /afx /b/bf/bfx /b/bg/bgx /c/cf/cfx /c/cg/cgx",
            "/a/af/afx?copy=/afx /a/af/afx /a/ag/agx /a/b/bf/bfx /a/b/bg/bgx /afx /b/bf/bfx /b/bg/bgx /c/cf/cfx /c/cg/cgx",
        ]

        self.conn = None
        self.fstab = None
        for dedup, act_exp in product(tc_dedup, tcs):
            action, expect = act_exp.split(" ", 1)
            t = "dedup:%s  action:%s" % (dedup, action)
            print("\n\n\033[0;7m# ", t, "\033[0m")

            ka = {"dav_inf": True}
            if dedup == "hard":
                ka["hardlink"] = True
            elif dedup == "no":
                ka["no_dedup"] = True

            self.args = Cfg(v=vols, a=[], **ka)
            self.reset()
            self.cinit()

            self.do_cp(action)
            zs = self.propfind()

            fns = " ".join(zs[1])
            self.assertEqual(expect, fns)

    def do_cp(self, action):
        hdr = "POST %s HTTP/1.1\r\nConnection: close\r\nContent-Length: 0\r\n\r\n"
        buf = (hdr % (action,)).encode("utf-8")
        print("CP [%s]" % (action,))
        HttpCli(self.conn.setbuf(buf)).run()
        ret = self.conn.s._reply.decode("utf-8").split("\r\n\r\n", 1)
        print("CP <-- ", ret)
        self.assertStart("HTTP/1.1 201 Created\r", ret[0])
        self.assertEqual("k\r\n", ret[1])
        return ret

    def propfind(self):
        h = "PROPFIND / HTTP/1.1\r\nConnection: close\r\n\r\n"
        HttpCli(self.conn.setbuf(h.encode("utf-8"))).run()
        h, t = self.conn.s._reply.decode("utf-8").split("\r\n\r\n", 1)
        fns = t.split("<D:response><D:href>")[1:]
        fns = [x.split("</D", 1)[0] for x in fns]
        fns = [x for x in fns if not x.endswith("/")]
        fns.sort()
        return h, fns

    def log(self, src, msg, c=0):
        print(msg)



# assume each requested file exists on another webserver and
# download + mirror them as they're requested
# (basically pretend we're warnish)

import os
import requests

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from copyparty.httpcli import HttpCli


def main(cli: "HttpCli", vn, rem):
    url = "https://mirrors.edge.kernel.org/alpine/" + rem
    abspath = os.path.join(vn.realpath, rem)

    # sneaky trick to preserve a requests-session between downloads
    # so it doesn't have to spend ages reopening https connections;
    # luckily we can stash it inside the copyparty client session,
    # name just has to be definitely unused so "hacapo_req_s" it is
    req_s = getattr(cli.conn, "hacapo_req_s", None) or requests.Session()
    setattr(cli.conn, "hacapo_req_s", req_s)

    try:
        os.makedirs(os.path.dirname(abspath), exist_ok=True)
        with req_s.get(url, stream=True, timeout=69) as r:
            r.raise_for_status()
            with open(abspath, "wb", 64 * 1024) as f:
                for buf in r.iter_content(chunk_size=64 * 1024):
                    f.write(buf)
    except:
        os.unlink(abspath)
        return "false"

    return "retry"



#!/usr/bin/env python3
# coding: utf-8
from __future__ import print_function, unicode_literals

import json
import os
import shutil
import tempfile
import unittest

from copyparty import util
from copyparty.authsrv import VFS, AuthSrv
from tests import util as tu
from tests.util import Cfg


class TestVFS(unittest.TestCase):
    def setUp(self):
        self.td = tu.get_ramdisk()

    def tearDown(self):
        os.chdir(tempfile.gettempdir())
        shutil.rmtree(self.td)

    def dump(self, vfs):
        print(json.dumps(vfs, indent=4, sort_keys=True, default=lambda o: o.__dict__))

    def unfoo(self, foo):
        for k, v in {"foo": "a", "bar": "b", "baz": "c", "qux": "d"}.items():
            foo = foo.replace(k, v)

        return foo

    def undot(self, vfs, query, response):
        self.assertEqual(util.undot(query), response)
        query = self.unfoo(query)
        response = self.unfoo(response)
        self.assertEqual(util.undot(query), response)

    def ls(self, vfs, vpath, uname):
        # type: (VFS, str, str) -> tuple[str, str, str]
        """helper for resolving and listing a folder"""
        vn, rem = vfs.get(vpath, uname, True, False)
        r1 = vn.ls(rem, uname, False, [[True]])
        r2 = vn.ls(rem, uname, False, [[True]])
        self.assertEqual(r1, r2)

        fsdir, real, virt = r1
        real = [x[0] for x in real]
        return fsdir, real, virt

    def log(self, src, msg, c=0):
        pass

    def assertAxs(self, dct, lst):
        t1 = list(sorted(dct))
        t2 = list(sorted(lst))
        self.assertEqual(t1, t2)

    def wipe_vfs(self, td):
        os.chdir("..")
        if os.path.exists(td):
            shutil.rmtree(td)
        os.mkdir(td)
        os.chdir(td)

        for a in "abc":
            for b in "abc":
                for c in "abc":
                    folder = "{0}/{0}{1}/{0}{1}{2}".format(a, b, c)
                    os.makedirs(folder)
                    for d in "abc":
                        fn = "{}/{}{}{}{}".format(folder, a, b, c, d)
                        with open(fn, "w") as f:
                            f.write(fn)

    def test(self):
        td = os.path.join(self.td, "vfs")
        self.wipe_vfs(td)

        # defaults
        self.wipe_vfs(td)
        vfs = AuthSrv(Cfg(), self.log).vfs
        self.assertEqual(vfs.nodes, {})
        self.assertEqual(vfs.vpath, "")
        self.assertEqual(vfs.realpath, td)
        self.assertAxs(vfs.axs.uread, ["*"])
        self.assertAxs(vfs.axs.uwrite, ["*"])

        # single read-only rootfs (relative path)
        self.wipe_vfs(td)
        vfs = AuthSrv(Cfg(v=["a/ab/::r"]), self.log).vfs
        self.assertEqual(vfs.nodes, {})
        self.assertEqual(vfs.vpath, "")
        self.assertEqual(vfs.realpath, os.path.join(td, "a", "ab"))
        self.assertAxs(vfs.axs.uread, ["*"])
        self.assertAxs(vfs.axs.uwrite, [])

        # single read-only rootfs (absolute path)
        self.wipe_vfs(td)
        vfs = AuthSrv(Cfg(v=[td + "//a/ac/../aa//::r"]), self.log).vfs
        self.assertEqual(vfs.nodes, {})
        self.assertEqual(vfs.vpath, "")
        self.assertEqual(vfs.realpath, os.path.join(td, "a", "aa"))
        self.assertAxs(vfs.axs.uread, ["*"])
        self.assertAxs(vfs.axs.uwrite, [])

        # read-only rootfs with write-only subdirectory (read-write for k)
        self.wipe_vfs(td)
        vfs = AuthSrv(
            Cfg(a=["k:k"], v=[".::r:rw,k", "a/ac/acb:a/ac/acb:w:rw,k"]),
            self.log,
        ).vfs
        self.assertEqual(len(vfs.nodes), 1)
        self.assertEqual(vfs.vpath, "")
        self.assertEqual(vfs.realpath, td)
        self.assertAxs(vfs.axs.uread, ["*", "k"])
        self.assertAxs(vfs.axs.uwrite, ["k"])
        n = vfs.nodes["a"]
        self.assertEqual(len(vfs.nodes), 1)
        self.assertEqual(n.vpath, "a")
        self.assertEqual(n.realpath, os.path.join(td, "a"))
        self.assertAxs(n.axs.uread, ["*", "k"])
        self.assertAxs(n.axs.uwrite, ["k"])
        n = n.nodes["ac"]
        self.assertEqual(len(vfs.nodes), 1)
        self.assertEqual(n.vpath, "a/ac")
        self.assertEqual(n.realpath, os.path.join(td, "a", "ac"))
        self.assertAxs(n.axs.uread, ["*", "k"])
        self.assertAxs(n.axs.uwrite, ["k"])
        n = n.nodes["acb"]
        self.assertEqual(n.nodes, {})
        self.assertEqual(n.vpath, "a/ac/acb")
        self.assertEqual(n.realpath, os.path.join(td, "a", "ac", "acb"))
        self.assertAxs(n.axs.uread, ["k"])
        self.assertAxs(n.axs.uwrite, ["*", "k"])

        # something funky about the windows path normalization,
        # doesn't really matter but makes the test messy, TODO?

        fsdir, real, virt = self.ls(vfs, "/", "*")
        self.assertEqual(fsdir, td)
        self.assertEqual(real, ["b", "c"])
        self.assertEqual(list(virt), ["a"])

        fsdir, real, virt = self.ls(vfs, "a", "*")
        self.assertEqual(fsdir, os.path.join(td, "a"))
        self.assertEqual(real, ["aa", "ab"])
        self.assertEqual(list(virt), ["ac"])

        fsdir, real, virt = self.ls(vfs, "a/ab", "*")
        self.assertEqual(fsdir, os.path.join(td, "a", "ab"))
        self.assertEqual(real, ["aba", "abb", "abc"])
        self.assertEqual(list(virt), [])

        fsdir, real, virt = self.ls(vfs, "a/ac", "*")
        self.assertEqual(fsdir, os.path.join(td, "a", "ac"))
        self.assertEqual(real, ["aca", "acc"])
        self.assertEqual(list(virt), [])

        fsdir, real, virt = self.ls(vfs, "a/ac", "k")
        self.assertEqual(fsdir, os.path.join(td, "a", "ac"))
        self.assertEqual(real, ["aca", "acc"])
        self.assertEqual(list(virt), ["acb"])

        self.assertRaises(util.Pebkac, vfs.get, "a/ac/acb", "*", True, False)

        fsdir, real, virt = self.ls(vfs, "a/ac/acb", "k")
        self.assertEqual(fsdir, os.path.join(td, "a", "ac", "acb"))
        self.assertEqual(real, ["acba", "acbb", "acbc"])
        self.assertEqual(list(virt), [])

        # admin-only rootfs with all-read-only subfolder
        self.wipe_vfs(td)
        vfs = AuthSrv(
            Cfg(a=["k:k"], v=[".::rw,k", "a:a:r"]),
            self.log,
        ).vfs
        self.assertEqual(len(vfs.nodes), 1)
        self.assertEqual(vfs.vpath, "")
        self.assertEqual(vfs.realpath, td)
        self.assertAxs(vfs.axs.uread, ["k"])
        self.assertAxs(vfs.axs.uwrite, ["k"])
        n = vfs.nodes["a"]
        self.assertEqual(len(vfs.nodes), 1)
        self.assertEqual(n.vpath, "a")
        self.assertEqual(n.realpath, os.path.join(td, "a"))
        self.assertAxs(n.axs.uread, ["*", "k"])
        self.assertAxs(n.axs.uwrite, [])
        perm_na = (False, False, False, False, False, False, False, False, False)
        perm_rw = (True, True, False, False, True, False, False, False, False)
        perm_ro = (True, False, False, False, True, False, False, False, False)
        self.assertEqual(vfs.can_access("/", "*"), perm_na)
        self.assertEqual(vfs.can_access("/", "k"), perm_rw)
        self.assertEqual(vfs.can_access("/a", "*"), perm_ro)
        self.assertEqual(vfs.can_access("/a", "k"), perm_ro)

        # breadth-first construction
        self.wipe_vfs(td)
        vfs = AuthSrv(
            Cfg(
                v=[
                    "a/ac/acb:a/ac/acb:w",
                    "a:a:w",
                    ".::r",
                    "abacdfasdq:abacdfasdq:w",
                    "a/ac:a/ac:w",
                ],
            ),
            self.log,
        ).vfs

        # sanitizing relative paths
        self.undot(vfs, "foo/bar/../baz/qux", "foo/baz/qux")
        self.undot(vfs, "foo/../bar", "bar")
        self.undot(vfs, "foo/../../bar", "bar")
        self.undot(vfs, "foo/../../", "")
        self.undot(vfs, "./.././foo/", "foo")
        self.undot(vfs, "./.././foo/..", "")

        # shadowing
        self.wipe_vfs(td)
        vfs = AuthSrv(Cfg(v=[".::r", "b:a/ac:r"]), self.log).vfs

        fsp, r1, v1 = self.ls(vfs, "", "*")
        self.assertEqual(fsp, td)
        self.assertEqual(r1, ["b", "c"])
        self.assertEqual(list(v1), ["a"])

        fsp, r1, v1 = self.ls(vfs, "a", "*")
        self.assertEqual(fsp, os.path.join(td, "a"))
        self.assertEqual(r1, ["aa", "ab"])
        self.assertEqual(list(v1), ["ac"])

        fsp1, r1, v1 = self.ls(vfs, "a/ac", "*")
        fsp2, r2, v2 = self.ls(vfs, "b", "*")
        self.assertEqual(fsp1, os.path.join(td, "b"))
        self.assertEqual(fsp2, os.path.join(td, "b"))
        self.assertEqual(r1, ["ba", "bb", "bc"])
        self.assertEqual(r1, r2)
        self.assertEqual(list(v1), list(v2))

        # config file parser
        cfg_path = os.path.join(self.td, "test.cfg")
        with open(cfg_path, "wb") as f:
            f.write(
                util.dedent(
                    """
                    u a:123
                    u asd:fgh:jkl

                    ./src
                    /dst
                    r a
                    rw asd
                    """
                ).encode("utf-8")
            )

        self.wipe_vfs(td)
        au = AuthSrv(Cfg(c=[cfg_path]), self.log)
        self.assertEqual(au.acct["a"], "123")
        self.assertEqual(au.acct["asd"], "fgh:jkl")
        n = au.vfs
        # root was not defined, so PWD with no access to anyone
        self.assertEqual(n.vpath, "")
        self.assertEqual(n.realpath, "")
        self.assertAxs(n.axs.uread, [])
        self.assertAxs(n.axs.uwrite, [])
        self.assertEqual(len(n.nodes), 1)
        n = n.nodes["dst"]
        self.assertEqual(n.vpath, "dst")
        self.assertEqual(n.realpath, os.path.join(td, "src"))
        self.assertAxs(n.axs.uread, ["a", "asd"])
        self.assertAxs(n.axs.uwrite, ["asd"])
        self.assertEqual(len(n.nodes), 0)

        os.unlink(cfg_path)



#!/usr/bin/env python3

import os
import sys
import subprocess as sp


"""
mtp test -- opens a texteditor

usage:
  -vsrv/v1:v1:r:c,mte=+x1:c,mtp=x1=ad,p,bin/mtag/mousepad.py

explained:
  c,mte: list of tags to index in this volume
  c,mtp: add new tag provider
     x1: dummy tag to provide
     ad: dontcare if audio or not
      p: priority 1 (run after initial tag-scan with ffprobe or mutagen)
"""


def main():
    env = os.environ.copy()
    env["DISPLAY"] = ":0.0"

    if False:
        # open the uploaded file
        fp = sys.argv[-1]
    else:
        # display stdin contents (`oth_tags`)
        fp = "/dev/stdin"

    p = sp.Popen(["/usr/bin/mousepad", fp])
    p.communicate()


main()



#!/usr/bin/env python3

import os
import re
import socket
import subprocess as sp


# to convert the copyparty --help to html, run this in xfce4-terminal @ 140x43:
_ = r""""
echo; for a in '' -bind -accounts -auth -auth-ord -flags -handlers -hooks -idp -urlform -exp -ls -dbd -chmod -pwhash -zm; do
./copyparty-sfx.py --help$a 2>/dev/null; printf '\n\n\n%0139d\n\n\n'; done  # xfce4-terminal @ 140x43
"""
# click [edit] => [select all]
# click [edit] => [copy as html]
# and then run this script


def readclip():
    cmds = [
        "xsel -ob",
        "xclip -selection CLIPBOARD -o",
        "pbpaste",
    ]
    if os.getenv("HLPTXT") == "CAT":
        cmds = ["cat"]

    for cmd in cmds:
        try:
            return sp.check_output(cmd.split()).decode("utf-8")
        except:
            pass
    raise Exception("need one of these: xsel xclip pbpaste")


def cnv(src):
    hostname = str(socket.gethostname()).split(".")[0]

    yield '<!DOCTYPE html>'
    yield '<html><body><style>'
    yield 'html{background:#222;color:#fff;line-height:1.25em}'
    yield 'h3{margin:0;padding:0}'
    yield 'a{color:#fc5;text-decoration:none;scroll-margin-top:3em}'
    yield 'a:active,a:target{background:#fc5;color:#000;box-shadow:0 0 0 .12em #fc5}'
    yield '</style>'
    skip_sfx = False
    in_sfx = 0
    in_salt = 0
    in_name = 0
    in_cores = 0
    in_hash_mt = False
    in_th_ram_max = 0

    while True:
        ln = next(src)
        if "<font" in ln or "<span" in ln:
            if not ln.startswith("<pre>"):
                ln = "<pre>" + ln
            yield ln
            break

    for ln in src:
        ln = ln.rstrip()
        t = ln
        if re.search(r"^<(font|span)[^>]+>copyparty v[0-9]", ln):
            in_sfx = 3
        if in_sfx:
            in_sfx -= 1
            if not skip_sfx:
                yield ln
            elif not in_sfx:
                yield "<span>"
            continue
        if '">uuid:' in ln:
            ln = re.sub(r">uuid:[0-9a-f-]{36}<", ">autogenerated<", ln)
        if "-salt SALT" in ln:
            in_salt = 3
        if in_salt:
            in_salt -= 1
            ln = re.sub(r">[0-9a-zA-Z/+]{24}<", ">24-character-autogenerated<", ln)
            ln = re.sub(r">[0-9a-zA-Z/+]{40}<", ">40-character-autogenerated<", ln)
            if t != ln:
                in_salt = 0
        if "--name TXT" in ln:
            in_name = 3
        if in_name:
            in_name -= 1
            ln = ln.replace(">" + hostname + "<", ">hostname<")
            if t != ln:
                in_name = 0
        if "--hash-mt CORES" in ln:
            in_cores = 3
            in_hash_mt = True
        if "--mtag-mt CORES" in ln or "--th-mt CORES" in ln:
            in_cores = 3
        if in_cores:
            in_cores -= 1
            zs = ">numCores"
            if in_hash_mt:
                zs += " if 5 or less"
            ln = re.sub(r">[0-9]{1,2}<", zs + "<", ln)
            if t != ln:
                in_cores = 0
                in_hash_mt = False
        if "--th-ram-max GB" in ln:
            in_th_ram_max = 3
        if in_th_ram_max:
            in_th_ram_max -= 1
            ln = re.sub(r">[0-9]{1,2}\.[0-9]<", ">dynamic<", ln)
            if t != ln:
                in_th_ram_max = 0
        m = re.search(r"^# (.* help page)(.*)", ln)
        if m:
            zs1, zs2 = m.groups()
            zs3 = zs1.replace(" ", "-")
            ln = '<h3># <a id="%s" href="#%s">%s</a>%s</h3></a>' % (zs3, zs3, zs1, zs2)
        m = re.search(r"^  (-{1,2})([^ ,]+)(.*)", ln)
        if m:
            zs1, zs2, zs3 = m.groups()
            ln = '  <a href="#g-%s" id="g-%s">%s%s</a>%s' % (zs2, zs2, zs1, zs2, zs3)

        ln = ln.replace(">/home/ed/", ">~/")
        if ln.startswith("0" * 20):
            skip_sfx = True
        yield ln

    yield "</pre>eof</body></html>"


def main():
    src = readclip()
    src = re.split("0{100,200}", src[::-1], maxsplit=1)[1][::-1]
    with open("helptext.html", "wb") as fo:
        for ln in cnv(iter(src.split("\n")[:-3])):
            fo.write(ln.encode("utf-8") + b"\r\n")


if __name__ == "__main__":
    main()



# sends a custom response instead of the usual 404


def main(cli, vn, rem):
    msg = f"sorry {cli.ip} but {cli.vpath} doesn't exist"

    return str(cli.reply(msg.encode("utf-8"), 404, "text/plain"))



#!/usr/bin/env python3
# coding: utf-8
from __future__ import print_function, unicode_literals

import unittest

from copyparty.__main__ import PY2
from copyparty.util import w8enc
from tests import util as tu


class TestUtils(unittest.TestCase):
    def cmp(self, orig, t1, t2):
        if t1 != t2:
            raise Exception("\n%r\n%r\n%r\n" % (w8enc(orig), t1, t2))

    def test_quotep(self):
        if PY2:
            raise unittest.SkipTest()

        from copyparty.util import _quotep3, _quotep3b, w8dec

        txt = w8dec(tu.randbytes(8192))
        self.cmp(txt, _quotep3(txt), _quotep3b(txt))

    def test_unquote(self):
        if PY2:
            raise unittest.SkipTest()

        from urllib.parse import unquote_to_bytes as u2b

        from copyparty.util import unquote

        for btxt in (
            tu.randbytes(8192),
            br"%ed%91qw,er;ty%20as df?gh+jkl%zxc&vbn <qwe>\"rty'uio&asd&nbsp;fgh",
        ):
            self.cmp(btxt, unquote(btxt), u2b(btxt))



#!/usr/bin/env python

import sys
import time
import json
import pefile

"""
retrieve exe info,
example for multivalue providers
"""


def unk(v):
    return "unk({:04x})".format(v)


class PE2(pefile.PE):
    def __init__(self, *a, **ka):
        for k in [
            # -- parse_data_directories:
            "parse_import_directory",
            "parse_export_directory",
            # "parse_resources_directory",
            "parse_debug_directory",
            "parse_relocations_directory",
            "parse_directory_tls",
            "parse_directory_load_config",
            "parse_delay_import_directory",
            "parse_directory_bound_imports",
            # -- full_load:
            "parse_rich_header",
        ]:
            setattr(self, k, self.noop)

        super(PE2, self).__init__(*a, **ka)

    def noop(*a, **ka):
        pass


try:
    pe = PE2(sys.argv[1], fast_load=False)
except:
    sys.exit(0)

arch = pe.FILE_HEADER.Machine
if arch == 0x14C:
    arch = "x86"
elif arch == 0x8664:
    arch = "x64"
else:
    arch = unk(arch)

try:
    buildtime = time.gmtime(pe.FILE_HEADER.TimeDateStamp)
    buildtime = time.strftime("%Y-%m-%d_%H:%M:%S", buildtime)
except:
    buildtime = "invalid"

ui = pe.OPTIONAL_HEADER.Subsystem
if ui == 2:
    ui = "GUI"
elif ui == 3:
    ui = "cmdline"
else:
    ui = unk(ui)

extra = {}
if hasattr(pe, "FileInfo"):
    for v1 in pe.FileInfo:
        for v2 in v1:
            if v2.name != "StringFileInfo":
                continue

            for v3 in v2.StringTable:
                for k, v in v3.entries.items():
                    v = v.decode("utf-8", "replace").strip()
                    if not v:
                        continue

                    if k in [b"FileVersion", b"ProductVersion"]:
                        extra["ver"] = v

                    if k in [b"OriginalFilename", b"InternalName"]:
                        extra["orig"] = v

r = {
    "arch": arch,
    "built": buildtime,
    "ui": ui,
    "cksum": "{:08x}".format(pe.OPTIONAL_HEADER.CheckSum),
}
r.update(extra)

print(json.dumps(r, indent=4))



#!/usr/bin/env python3

import re, os, sys, codecs

outfile = os.path.realpath(sys.argv[1])

os.chdir(os.path.dirname(__file__))

with open("../docs/lics.txt", "rb") as f:
    s = f.read().decode("utf-8").rstrip("\n") + "\n\n\n\n"
    s = re.sub("\nC: ", "\nCopyright (c) ", s)
    s = re.sub("\nL: ", "\nLicense: ", s)
    ret = s.split("\n")

lics = [
    "MIT License",
    "BSD 2-Clause License",
    "BSD 3-Clause License",
    "ISC License",
    "Apache License v2.0",
    "SIL Open Font License v1.1",
]

for n, lic in enumerate(lics, 1):
    with open("lics/%d.r13" % (n,), "rb") as f:
        s = f.read().decode("utf-8")
        s = codecs.decode(s, "rot_13")
        s = "\n--- %s ---\n\n%s" % (lic, s)
        ret.extend(s.split("\n"))

for n, ln in enumerate(ret):
    if not ln.startswith("--- "):
        continue
    pad = " " * ((80 - len(ln)) // 2)
    ln = "%s\033[07m%s\033[0m" % (pad, ln)
    ret[n] = ln

ret.append("")
ret.append("")

with open(outfile, "wb") as f:
    f.write(("\n".join(ret)).encode("utf-8"))



#!/usr/bin/env python3
# coding: utf-8
from __future__ import print_function, unicode_literals

import os
import shutil
import tempfile
import time
import unittest

from copyparty.__init__ import MACOS
from copyparty.authsrv import AuthSrv
from copyparty.httpcli import HttpCli
from tests import util as tu
from tests.util import TC, Cfg, pfind2ls

# tcpdump of `rclone ls dav:`
RCLONE_PROPFIND = """PROPFIND /%s HTTP/1.1
Host: 127.0.0.1:3923
User-Agent: rclone/v1.67.0
Content-Length: 308
Authorization: Basic azp1
Depth: 1
Referer: http://127.0.0.1:3923/
Accept-Encoding: gzip

<?xml version="1.0"?>
<d:propfind  xmlns:d="DAV:" xmlns:oc="http://owncloud.org/ns" xmlns:nc="http://nextcloud.org/ns">
 <d:prop>
  <d:displayname />
  <d:getlastmodified />
  <d:getcontentlength />
  <d:resourcetype />
  <d:getcontenttype />
  <oc:checksums />
  <oc:permissions />
 </d:prop>
</d:propfind>
"""


# tcpdump of `rclone copy fa dav:/a/`  (it does a mkcol first)
RCLONE_MKCOL = """MKCOL /%s HTTP/1.1
Host: 127.0.0.1:3923
User-Agent: rclone/v1.67.0
Authorization: Basic azp1
Referer: http://127.0.0.1:3923/
Accept-Encoding: gzip
\n"""


# tcpdump of `rclone copy fa dav:/a/`  (the actual upload)
RCLONE_PUT = """PUT /%s HTTP/1.1
Host: 127.0.0.1:3923
User-Agent: rclone/v1.67.0
Content-Length: 6
Authorization: Basic azp1
Content-Type: application/octet-stream
Oc-Checksum: SHA1:f5e3dc3fb27af53cd0005a1184e2df06481199e8
Referer: http://127.0.0.1:3923/
X-Oc-Mtime: 1689453578
Accept-Encoding: gzip

fgsfds"""


RCLONE_PUT_FLOAT = """PUT /%s HTTP/1.1
Host: 127.0.0.1:3923
User-Agent: rclone/v1.67.0
Content-Length: 6
Authorization: Basic azp1
Content-Type: application/octet-stream
Oc-Checksum: SHA1:f5e3dc3fb27af53cd0005a1184e2df06481199e8
Referer: http://127.0.0.1:3923/
X-Oc-Mtime: 1689453578.123
Accept-Encoding: gzip

fgsfds"""


# tcpdump of `rclone delete dav:/a/d1/`  (it does propfind recursively and then this on each file)
# (note: `rclone rmdirs dav:/a/d1/` does the same thing but just each folder after asserting they're empty)
RCLONE_DELETE = """DELETE /%s HTTP/1.1
Host: 127.0.0.1:3923
User-Agent: rclone/v1.67.0
Authorization: Basic azp1
Referer: http://127.0.0.1:3923/
Accept-Encoding: gzip
\n"""


# tcpdump of `rclone move dav:/a/d1/d2 /a/d1/d3`  (it does a lot of boilerplate propfinds/mkcols before)
RCLONE_MOVE = """MOVE /%s HTTP/1.1
Host: 127.0.0.1:3923
User-Agent: rclone/v1.67.0
Authorization: Basic azp1
Destination: http://127.0.0.1:3923/%s
Overwrite: T
Referer: http://127.0.0.1:3923/
Accept-Encoding: gzip
\n"""


class TestHttpCli(TC):
    def setUp(self):
        self.td = tu.get_ramdisk()
        self.maxDiff = 99999

    def tearDown(self):
        self.conn.shutdown()
        os.chdir(tempfile.gettempdir())
        shutil.rmtree(self.td)

    def test(self):
        td = os.path.join(self.td, "vfs")
        os.mkdir(td)
        os.chdir(td)

        self.fn = "g{:x}g".format(int(time.time() * 3))
        vcfg = [
            "r:r:r,u",
            "w:w:w,u",
            "a:a:A,u",
            "x:x:r,u2",
            "x/r:x/r:r,u",
            "x/x:x/x:r,u2",
        ]
        self.args = Cfg(v=vcfg, a=["u:u", "u2:u2"])
        self.asrv = AuthSrv(self.args, self.log)
        self.conn = tu.VHttpConn(self.args, self.asrv, self.log, b"", True)

        self.fns = ["%s/%s" % (zs.split(":")[0], self.fn) for zs in vcfg]
        for fp in self.fns:
            try:
                os.makedirs(os.path.dirname(fp))
            except:
                pass
            with open(fp, "wb") as f:
                f.write(("ok %s\n" % (fp,)).encode("utf-8"))

        ##
        ## depth:1 (regular listing)

        # unmapped root; should return list of volumes
        h, b = self.req(RCLONE_PROPFIND % ("",))
        fns = pfind2ls(b)
        self.assertStart("HTTP/1.1 207 Multi-Status\r", h)
        self.assertListEqual(fns, ["/", "/a/", "/r/"])

        # toplevel of a volume; has one file
        h, b = self.req(RCLONE_PROPFIND % ("a",))
        fns = pfind2ls(b)
        self.assertStart("HTTP/1.1 207 Multi-Status\r", h)
        self.assertListEqual(fns, ["/a/", "/a/" + self.fn])

        # toplevel of a volume; has one file
        h, b = self.req(RCLONE_PROPFIND % ("r",))
        fns = pfind2ls(b)
        self.assertStart("HTTP/1.1 207 Multi-Status\r", h)
        self.assertListEqual(fns, ["/r/", "/r/" + self.fn])

        # toplevel of write-only volume; has one file, will not list
        h, b = self.req(RCLONE_PROPFIND % ("w",))
        fns = pfind2ls(b)
        self.assertStart("HTTP/1.1 207 Multi-Status\r", h)
        self.assertListEqual(fns, ["/w/"])

        ##
        ## auth challenge

        bad_pfind = RCLONE_PROPFIND.replace("Authorization: Basic azp1\n", "")
        bad_put = RCLONE_PUT.replace("Authorization: Basic azp1\n", "")
        urls = ["", "r", "w", "a"]
        urls += [x + "/" + self.fn for x in urls[1:]]
        for url in urls:
            for q in (bad_pfind, bad_put):
                h, b = self.req(q % (url,))
                self.assertStart("HTTP/1.1 401 Unauthorized\r", h)
                self.assertIn('\nWWW-Authenticate: Basic realm="a"\r', h)

        ##
        ## depth:0 (recursion)

        # depth:0 from unmapped root should work;
        # will NOT list contents of /x/r/ due to current limitations
        # (stops descending at first non-accessible volume)
        recursive = RCLONE_PROPFIND.replace("Depth: 1\n", "")
        h, b = self.req(recursive % ("",))
        fns = pfind2ls(b)
        expect = ["/", "/a/", "/r/"]
        expect += [x + self.fn for x in expect[1:]]
        self.assertListEqual(fns, expect)

        # same thing here...
        h, b = self.req(recursive % ("/x",))
        fns = pfind2ls(b)
        self.assertListEqual(fns, [])

        # but this obviously works
        h, b = self.req(recursive % ("/x/r",))
        fns = pfind2ls(b)
        self.assertListEqual(fns, ["/x/r/", "/x/r/" + self.fn])

        ##
        ## uploading

        # rclone does a propfind on the target file first; expects 404
        h, b = self.req(RCLONE_PROPFIND % ("a/fa",))
        self.assertStart("HTTP/1.1 404 Not Found\r", h)

        # then it does a mkcol (mkdir), expecting 405 (exists)
        h, b = self.req(RCLONE_MKCOL % ("a",))
        self.assertStart("HTTP/1.1 405 Method Not Allowed\r", h)

        # then it uploads the file
        h, b = self.req(RCLONE_PUT % ("a/fa",))
        self.assertStart("HTTP/1.1 201 Created\r", h)

        # float x-oc-mtime should be accepted
        h, b = self.req(RCLONE_PUT_FLOAT % ("a/fb",))
        self.assertStart("HTTP/1.1 201 Created\r", h)
        zi = 0 if MACOS else 3  # its okay macos you tried your best
        self.assertAlmostEqual(os.path.getmtime("a/fb"), 1689453578.123, places=zi)

        # then it does a propfind to confirm
        h, b = self.req(RCLONE_PROPFIND % ("a/fa",))
        fns = pfind2ls(b)
        self.assertStart("HTTP/1.1 207 Multi-Status\r", h)
        self.assertListEqual(fns, ["/a/fa"])

        ##
        ## upload into set of subfolders that don't exist yet

        # rclone does this:
        # propfind /a/d1/d2/fa => 404
        # mkcol /a/d1/d2/ => 409
        # propfind /a/d1/d2/ => 404
        # mkcol /a/d1/ => 201
        # mkcol /a/d1/d2/ => 201
        # put /a/d1/d2/fa => 201
        # propfind /a/d1/d2/fa => 207
        # ...some of which already tested above;

        h, b = self.req(RCLONE_PROPFIND % ("/a/d1/d2/",))
        self.assertStart("HTTP/1.1 404 Not Found\r", h)

        h, b = self.req(RCLONE_PROPFIND % ("/a/d1/",))
        self.assertStart("HTTP/1.1 404 Not Found\r", h)

        h, b = self.req(RCLONE_MKCOL % ("/a/d1/d2/",))
        self.assertStart("HTTP/1.1 409 Conflict\r", h)

        h, b = self.req(RCLONE_MKCOL % ("/a/d1/",))
        self.assertStart("HTTP/1.1 201 Created\r", h)

        h, b = self.req(RCLONE_MKCOL % ("/a/d1/d2/",))
        self.assertStart("HTTP/1.1 201 Created\r", h)

        h, b = self.req(RCLONE_PUT % ("a/d1/d2/fa",))
        self.assertStart("HTTP/1.1 201 Created\r", h)

        ##
        ## rename

        h, b = self.req(RCLONE_MOVE % ("a/d1/d2/", "a/d1/d3/"))
        self.assertStart("HTTP/1.1 201 Created\r", h)
        self.assertListEqual(os.listdir("a/d1"), ["d3"])

        ##
        ## delete

        h, b = self.req(RCLONE_DELETE % ("a/d1",))
        self.assertStart("HTTP/1.1 200 OK\r", h)
        if os.path.exists("a/d1"):
            self.fail("a/d1 still exists")

    def req(self, q):
        h, b = q.split("\n\n", 1)
        q = h.replace("\n", "\r\n") + "\r\n\r\n" + b
        conn = self.conn.setbuf(q.encode("utf-8"))
        HttpCli(conn).run()
        return conn.s._reply.decode("utf-8").split("\r\n\r\n", 1)

    def log(self, src, msg, c=0):
        print(msg)



#!/usr/bin/env python3

import os
import time

"""
td=/dev/shm/; [ -e $td ] || td=$HOME; mkdir -p $td/fusefuzz/{r,v}
PYTHONPATH=.. python3 -m copyparty -v $td/fusefuzz/r::r -i 127.0.0.1
../bin/partyfuse.py http://127.0.0.1:3923/ $td/fusefuzz/v -cf 2 -cd 0.5
(d="$PWD"; cd $td/fusefuzz && "$d"/fusefuzz.py)
"""


def chk(fsz, rsz, ofs0, shift, ofs, rf, vf):
    if ofs != rf.tell():
        rf.seek(ofs)
        vf.seek(ofs)

    rb = rf.read(rsz)
    vb = vf.read(rsz)

    print(f"fsz {fsz} rsz {rsz} ofs {ofs0} shift {shift} ofs {ofs} = {len(rb)}")

    if rb != vb:
        for n, buf in enumerate([rb, vb]):
            with open("buf." + str(n), "wb") as f:
                f.write(buf)

        raise Exception(f"{len(rb)} != {len(vb)}")

    return rb, vb


def main():
    v = "v"
    for n in range(5):
        with open(f"r/{n}", "wb") as f:
            f.write(b"h" * n)

    rand = os.urandom(7919)  # prime
    for fsz in range(1024 * 1024 * 2 - 3, 1024 * 1024 * 2 + 3):
        with open("r/f", "wb", fsz) as f:
            f.write((rand * int(fsz / len(rand) + 1))[:fsz])

        for rsz in range(64 * 1024 - 2, 64 * 1024 + 2):
            ofslist = [0, 1, 2]
            for n in range(3):
                ofslist.append(fsz - n)
                ofslist.append(fsz - (rsz * 1 + n))
                ofslist.append(fsz - (rsz * 2 + n))

            for ofs0 in ofslist:
                for shift in range(-3, 3):
                    print(f"fsz {fsz} rsz {rsz} ofs {ofs0} shift {shift}")
                    ofs = ofs0
                    if ofs < 0 or ofs >= fsz:
                        continue

                    for n in range(1, 3):
                        with open(f"{v}/{n}", "rb") as f:
                            f.read()

                    prev_ofs = -99
                    with open("r/f", "rb", rsz) as rf:
                        with open(f"{v}/f", "rb", rsz) as vf:
                            while True:
                                ofs += shift
                                if ofs < 0 or ofs > fsz or ofs == prev_ofs:
                                    break

                                prev_ofs = ofs

                                rb, vb = chk(fsz, rsz, ofs0, shift, ofs, rf, vf)

                                if not rb:
                                    break

                                ofs += len(rb)

                    for n in range(1, 3):
                        with open(f"{v}/{n}", "rb") as f:
                            f.read()

                    with open("r/f", "rb", rsz) as rf:
                        with open(f"{v}/f", "rb", rsz) as vf:
                            for n in range(2):
                                ofs += shift
                                if ofs < 0 or ofs > fsz:
                                    break

                                rb, vb = chk(fsz, rsz, ofs0, shift, ofs, rf, vf)

                                ofs -= rsz

        # bumping fsz, sleep away the dentry cache in cppf
        time.sleep(1)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# coding: utf-8
from __future__ import print_function, unicode_literals

import io
import os
import sys
import tokenize


try:
    FSTRING_MIDDLE = tokenize.FSTRING_MIDDLE
except:
    FSTRING_MIDDLE = -9001


def uncomment(fpath):
    """modified https://stackoverflow.com/a/62074206"""

    print(".", end="", flush=True)
    with open(fpath, "rb") as f:
        orig = f.read().decode("utf-8")

    out = ""
    io_obj = io.StringIO(orig)
    prev_toktype = tokenize.INDENT
    last_lineno = -1
    last_col = 0
    code = False
    for tok in tokenize.generate_tokens(io_obj.readline):
        # print(repr(tok))
        token_type = tok[0]
        token_string = tok[1]
        start_line, start_col = tok[2]
        end_line, end_col = tok[3]

        if start_line > last_lineno:
            last_col = 0

        if start_col > last_col and prev_toktype != FSTRING_MIDDLE:
            out += " " * (start_col - last_col)

        is_legalese = (
            "copyright" in token_string.lower() or "license" in token_string.lower()
        )

        if token_type == tokenize.STRING:
            if (
                prev_toktype != tokenize.INDENT
                and prev_toktype != tokenize.NEWLINE
                and start_col > 0
                or is_legalese
            ):
                out += token_string
            else:
                out += '"a"'
        elif token_type == FSTRING_MIDDLE:
            out += token_string.replace(r"{", r"{{").replace(r"}", r"}}")
            if not code and token_string.strip():
                code = True
        elif token_type != tokenize.COMMENT:
            out += token_string
            if not code and token_string.strip():
                code = True
        elif is_legalese or (not start_col and not code):
            out += token_string
        else:
            if out.rstrip(" ").endswith("\n"):
                out = out.rstrip() + "\n"
            else:
                out = out.rstrip()

        prev_toktype = token_type
        last_lineno = end_line
        last_col = end_col

    # out = "\n".join(x for x in out.splitlines() if x.strip())

    with open(fpath, "wb") as f:
        f.write(out.encode("utf-8"))


def main():
    print("uncommenting", end="", flush=True)
    try:
        if sys.argv[1] == "1":
            sys.argv.remove("1")
            raise Exception("disabled")

        import multiprocessing as mp

        mp.set_start_method("spawn", True)
        with mp.Pool(os.cpu_count()) as pool:
            pool.map(uncomment, sys.argv[1:])
    except Exception as ex:
        print("\nnon-mp fallback due to {}\n".format(ex))
        for f in sys.argv[1:]:
            try:
                uncomment(f)
            except Exception as ex:
                print("uncomment failed: [%s] %s" % (f, repr(ex)))

    print("k")


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import os


_ = r"""
rename incoming uploads according to the "wark" (the file identifier)
which is basically but not exactly a sha512 hash of the file contents

NOTE: this only works for up2k uploads (dragdrop into browser);
  combine this with reloc-by-wark-xau.py to cover the other protocols

example usage as global config:
    -e2d --xbu I,c,bin/hooks/reloc-by-wark-xbu.py

parameters explained,
    e2d = enable up2k database (mandatory for xbu hooks)
    xbu = execute before upload
    I   = import this hook for performance; do not fork / subprocess
    c   = "check"; reject upload if this hook crashes due to a bug

example usage as a volflag (per-volume config):
    -v srv/inc:inc:r:rw,ed:c,e2d,xbu=I,c,bin/hooks/reloc-by-wark-xbu.py
                           ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

    (share filesystem-path srv/inc as volume /inc,
     readable by everyone, read-write for user 'ed',
     running this plugin on all uploads with the params explained above)

example usage as a volflag in a copyparty config file:
    [/inc]
      srv/inc
      accs:
        r: *
        rw: ed
      flags:
        e2d, xbu: I,c,bin/hooks/reloc-by-wark-xbu.py
"""


def main(inf):
    wark = inf.get("wark")
    if not wark:
        # not an up2k upload, so we don't have the hash;

        # option 1: let upload proceed with original filename
        return {}

        # option 2: reject the upload
        return {"rejectmsg": "only up2k uploads are allowed in this volume"}

    # grab the original filename from the vpath...
    vdir, fn = os.path.split(inf["vp"])

    # ...to retain the original file extension, if any
    try:
        fn, ext = fn.rsplit(".", 1)
    except:
        ext = ""

    # use the first 16 characters; 12 bytes of entropy,
    # roughly one collision for every 26 million files
    fn = wark[:16]

    if ext:
        ext = ext.lower()
        fn += "." + ext

    return {"reloc": {"fn": fn}}



#!/usr/bin/env python
# coding: utf-8
# vim:ts=4:sw=4:softtabstop=4:smarttab:expandtab
from __future__ import print_function, unicode_literals

import json
import os
import sys
import time

try:
    from datetime import datetime, timezone
except:
    from datetime import datetime


_ = r"""
use copyparty as a dumb messaging server / guestbook thing;
accepts guestbook entries from 📟 (message-to-server-log) in the web-ui
initially contributed by @clach04 in https://github.com/9001/copyparty/issues/35 (thanks!)

example usage as global config:
    python copyparty-sfx.py --xm j,bin/hooks/msg-log.py

parameters explained,
    xm = execute on message (📟)
    j  = this hook needs message information as json (not just the message-text)

example usage as a volflag (per-volume config):
    python copyparty-sfx.py -v srv/log:log:r:c,xm=j,bin/hooks/msg-log.py
                                             ^^^^^^^^^^^^^^^^^^^^^^^^^^^

    (share filesystem-path srv/log as volume /log, readable by everyone,
     running this plugin on all messages with the params explained above)

example usage as a volflag in a copyparty config file:
    [/log]
      srv/log
      accs:
        r: *
      flags:
        xm: j,bin/hooks/msg-log.py
"""


# output filename
FILENAME = os.environ.get("COPYPARTY_MESSAGE_FILENAME", "") or "README.md"

# set True to write in descending order (newest message at top of file);
# note that this becomes very slow/expensive as the file gets bigger
DESCENDING = True

# the message template; the following parameters are provided by copyparty and can be referenced below:
# 'ap' = absolute filesystem path where the message was posted
# 'vp' = virtual path (URL 'path') where the message was posted
# 'mt' = 'at' = unix-timestamp when the message was posted
# 'datetime' = ISO-8601 time when the message was posted
# 'sz' = message size in bytes
# 'host' = the server hostname which the user was accessing (URL 'host')
# 'user' = username (if logged in), otherwise '*'
# 'txt' = the message text itself
# (uncomment the print(msg_info) to see if additional information has been introduced by copyparty since this was written)
TEMPLATE = """
🕒 %(datetime)s, 👤 %(user)s @ %(ip)s
%(txt)s
"""


def write_ascending(filepath, msg_text):
    with open(filepath, "a", encoding="utf-8", errors="replace") as outfile:
        outfile.write(msg_text)


def write_descending(filepath, msg_text):
    lockpath = filepath + ".lock"
    got_it = False
    for _ in range(16):
        try:
            os.mkdir(lockpath)
            got_it = True
            break
        except:
            time.sleep(0.1)
            continue

    if not got_it:
        return sys.exit(1)

    try:
        oldpath = filepath + ".old"
        os.rename(filepath, oldpath)
        with open(oldpath, "r", encoding="utf-8", errors="replace") as infile, open(
            filepath, "w", encoding="utf-8", errors="replace"
        ) as outfile:
            outfile.write(msg_text)
            while True:
                buf = infile.read(4096)
                if not buf:
                    break
                outfile.write(buf)
    finally:
        try:
            os.unlink(oldpath)
        except:
            pass
        os.rmdir(lockpath)


def main(argv=None):
    if argv is None:
        argv = sys.argv

    msg_info = json.loads(sys.argv[1])
    # print(msg_info)

    try:
        dt = datetime.fromtimestamp(msg_info["at"], timezone.utc)
    except:
        dt = datetime.utcfromtimestamp(msg_info["at"])

    msg_info["datetime"] = dt.strftime("%Y-%m-%d, %H:%M:%S")

    msg_text = TEMPLATE % msg_info

    filepath = os.path.join(msg_info["ap"], FILENAME)

    if DESCENDING and os.path.exists(filepath):
        write_descending(filepath, msg_text)
    else:
        write_ascending(filepath, msg_text)

    print(msg_text)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# coding: utf-8
from __future__ import print_function, unicode_literals

import io
import json
import os
import shutil
import tarfile
import tempfile
import unittest

from copyparty.authsrv import AuthSrv
from copyparty.httpcli import HttpCli
from copyparty.u2idx import U2idx
from copyparty.up2k import Up2k
from tests import util as tu
from tests.util import J2_ENV, Cfg

try:
    from typing import Optional
except:
    pass


def hdr(query, uname, extra=""):
    h = "GET /%s HTTP/1.1\r\nPW: %s\r\nConnection: close\r\n%s\r\n"
    return (h % (query, uname, extra)).encode("utf-8")


class TestDots(unittest.TestCase):
    def __init__(self, *a, **ka):
        super(TestDots, self).__init__(*a, **ka)
        self.is_dut = True
        self.j2_brw = None

    def setUp(self):
        self.conn: Optional[tu.VHttpConn] = None
        self.td = tu.get_ramdisk()

    def tearDown(self):
        if self.conn:
            self.conn.shutdown()
        os.chdir(tempfile.gettempdir())
        shutil.rmtree(self.td)

    def cinit(self):
        if self.conn:
            self.conn.shutdown()
            self.conn = None
        self.conn = tu.VHttpConn(self.args, self.asrv, self.log, b"")

        if not self.j2_brw:
            zs = os.path.dirname(__file__)
            with open("%s/../copyparty/web/browser.html" % (zs,), "rb") as f:
                zs = f.read().decode("utf-8")
            self.j2_brw = J2_ENV.from_string(zs)
        self.conn.hsrv.j2["browser"] = self.j2_brw

    def test_dots(self):
        td = os.path.join(self.td, "vfs")
        os.mkdir(td)
        os.chdir(td)

        # topDir  volA  volA/*dirA  .volB  .volB/*dirB
        spaths = " t .t a a/da a/.da .b .b/db .b/.db"
        for n, dirpath in enumerate(spaths.split(" ")):
            if dirpath:
                os.makedirs(dirpath)

            for pfx in "f", ".f":
                filepath = pfx + str(n)
                if dirpath:
                    filepath = os.path.join(dirpath, filepath)

                with open(filepath, "wb") as f:
                    f.write(filepath.encode("utf-8"))

        vcfg = [".::r,u1:r.,u2", "a:a:r,u1:r,u2", ".b:.b:r.,u1:r,u2"]
        self.args = Cfg(v=vcfg, a=["u1:u1", "u2:u2"], e2dsa=True)
        self.asrv = AuthSrv(self.args, self.log)
        self.cinit()

        self.assertEqual(self.tardir("", "u1"), "f0 t/f1 a/f3 a/da/f4")
        self.assertEqual(self.tardir(".t", "u1"), "f2")
        self.assertEqual(self.tardir(".b", "u1"), ".f6 f6 .db/.f8 .db/f8 db/.f7 db/f7")

        zs = ".f0 f0 .t/.f2 .t/f2 t/.f1 t/f1 .b/f6 .b/db/f7 a/f3 a/da/f4"
        self.assertEqual(self.tardir("", "u2"), zs)

        self.assertEqual(self.curl("?tar", "x")[1][:17], "\nJ2EOT")

        ##
        ## search

        up2k = Up2k(self)
        u2idx = U2idx(self)
        allvols = list(self.asrv.vfs.all_vols.values())

        x = u2idx.search("u1", allvols, "", 999)
        x = " ".join(sorted([x["rp"] for x in x[0]]))
        # u1 can see dotfiles in volB so they should be included
        xe = ".b/.db/.f8 .b/.db/f8 .b/.f6 .b/db/.f7 .b/db/f7 .b/f6 a/da/f4 a/f3 f0 t/f1"
        self.assertEqual(x, xe)

        x = u2idx.search("u2", allvols, "", 999)
        x = " ".join(sorted([x["rp"] for x in x[0]]))
        self.assertEqual(x, ".f0 .t/.f2 .t/f2 a/da/f4 a/f3 f0 t/.f1 t/f1")

        u2idx.shutdown()
        self.args = Cfg(v=vcfg, a=["u1:u1", "u2:u2"], dotsrch=False, e2d=True)
        self.asrv = AuthSrv(self.args, self.log)
        u2idx = U2idx(self)
        self.cinit()

        x = u2idx.search("u1", self.asrv.vfs.all_vols.values(), "", 999)
        x = " ".join(sorted([x["rp"] for x in x[0]]))
        # u1 can see dotfiles in volB so they should be included
        xe = "a/da/f4 a/f3 f0 t/f1"
        self.assertEqual(x, xe)
        u2idx.shutdown()
        up2k.shutdown()

        ##
        ## dirkeys

        os.mkdir("v")
        with open("v/f1.txt", "wb") as f:
            f.write(b"a")
        os.rename("a", "v/a")
        os.rename(".b", "v/.b")

        vcfg = [
            ".::r.,u1:g,u2:c,dks",
            "v/a:v/a:r.,u1:g,u2:c,dks",
            "v/.b:v/.b:r.,u1:g,u2:c,dks",
        ]
        self.args = Cfg(v=vcfg, a=["u1:u1", "u2:u2"])
        self.asrv = AuthSrv(self.args, self.log)
        self.cinit()

        zj = json.loads(self.curl("?ls", "u1")[1])
        url = "?k=" + zj["dk"]
        # should descend into folders, but not other volumes:
        self.assertEqual(self.tardir(url, "u2"), "f0 t/f1 v/f1.txt")

        zj = json.loads(self.curl("v?ls", "u1")[1])
        url = "v?k=" + zj["dk"]
        self.assertEqual(self.tarsel(url, "u2", ["f1.txt", "a", ".b"]), "f1.txt")

        shutil.rmtree("v")

    def test_dk_fk(self):
        # python3 -m unittest tests.test_dots.TestDots.test_dk_fk

        td = os.path.join(self.td, "vfs")
        os.mkdir(td)
        os.chdir(td)

        vcfg = []
        zs = "dk dks=12 dky fk fka dk,fk dks=12:c,fk"
        for k2 in zs.split():
            k = k2.replace("=12", "").replace(":c,", ",")
            vcfg += ["{0}:{0}:r.,u1:g,u2:c,{1}".format(k, k2)]
            zs = "%s/s1/s2" % (k,)
            os.makedirs(zs)

            with open("%s/f.t1" % (k,), "wb") as f:
                f.write(b"f1")

            with open("%s/s1/f.t2" % (k,), "wb") as f:
                f.write(b"f2")

            with open("%s/s1/s2/f.t3" % (k,), "wb") as f:
                f.write(b"f3")

        self.args = Cfg(v=vcfg, a=["u1:u1", "u2:u2"])
        self.asrv = AuthSrv(self.args, self.log)
        self.cinit()

        dk = {}
        for d in "dk dks dk,fk dks,fk".split():
            zj = json.loads(self.curl("%s?ls" % (d,), "u1")[1])
            dk[d] = zj["dk"]

        ##
        ## dk

        # should not be able to access dk with wrong dirkey,
        zs = self.curl("dk?ls&k=%s" % (dk["dks"]), "u2")[1]
        self.assertEqual(zs, "\nJ2EOT")
        # so use the right key
        zs = self.curl("dk?ls&k=%s" % (dk["dk"]), "u2")[1]
        zj = json.loads(zs)
        self.assertEqual(len(zj["dirs"]), 0)
        self.assertEqual(len(zj["files"]), 1)
        self.assertEqual(zj["files"][0]["href"], "f.t1")

        ##
        ## dk thumbs

        self.assertIn('">folder</text>', self.curl("dk?th=x", "u1")[1])
        self.assertIn('">e403</text>', self.curl("dk?th=x", "u2")[1])

        zs = "dk?th=x&k=%s" % (dk["dks"])
        self.assertIn('">e403</text>', self.curl(zs, "u2")[1])

        zs = "dk?th=x&k=%s" % (dk["dk"])
        self.assertIn('">folder</text>', self.curl(zs, "u2")[1])

        # fk not enabled, so this should work
        self.assertIn('">t1</text>', self.curl("dk/f.t1?th=x", "u2")[1])
        self.assertIn('">t2</text>', self.curl("dk/s1/f.t2?th=x", "u2")[1])

        ##
        ## dks

        # should not be able to access dks with wrong dirkey,
        zs = self.curl("dks?ls&k=%s" % (dk["dk"]), "u2")[1]
        self.assertEqual(zs, "\nJ2EOT")
        # so use the right key
        zs = self.curl("dks?ls&k=%s" % (dk["dks"]), "u2")[1]
        zj = json.loads(zs)
        self.assertEqual(len(zj["dirs"]), 1)
        self.assertEqual(len(zj["files"]), 1)
        self.assertEqual(zj["files"][0]["href"], "f.t1")
        # dks should return correct dirkey of subfolders;
        s1 = zj["dirs"][0]["href"]
        self.assertEqual(s1.split("/")[0], "s1")
        zs = self.curl("dks/%s&ls" % (s1), "u2")[1]
        self.assertIn('"s2/?k=', zs)

        # parent key should not exist in response
        self.assertNotIn(dk["dks"], zs)
        # and not in html either
        for ck in ("", "Cookie: js=y\r\n"):
            zb = hdr("dks/%s" % (s1), "u2", ck)
            zs = self.curl("", "", False, zb)[1]
            self.assertNotIn(dk["dks"], zs)
            self.assertIn('"s2/?k=', zs)

        ##
        ## dks thumbs

        self.assertIn('">folder</text>', self.curl("dks?th=x", "u1")[1])
        self.assertIn('">e403</text>', self.curl("dks?th=x", "u2")[1])

        zs = "dks?th=x&k=%s" % (dk["dk"])
        self.assertIn('">e403</text>', self.curl(zs, "u2")[1])

        zs = "dks?th=x&k=%s" % (dk["dks"])
        self.assertIn('">folder</text>', self.curl(zs, "u2")[1])

        # fk not enabled, so this should work
        self.assertIn('">t1</text>', self.curl("dks/f.t1?th=x", "u2")[1])
        self.assertIn('">t2</text>', self.curl("dks/s1/f.t2?th=x", "u2")[1])

        ##
        ## dky

        # doesn't care about keys
        zs = self.curl("dky?ls&k=ok", "u2")[1]
        self.assertEqual(zs, self.curl("dky?ls", "u2")[1])
        zj = json.loads(zs)
        self.assertEqual(len(zj["dirs"]), 0)
        self.assertEqual(len(zj["files"]), 1)
        self.assertEqual(zj["files"][0]["href"], "f.t1")

        ##
        ## dky thumbs

        self.assertIn('">folder</text>', self.curl("dky?th=x", "u1")[1])
        self.assertIn('">folder</text>', self.curl("dky?th=x", "u2")[1])

        zs = "dky?th=x&k=%s" % (dk["dk"])
        self.assertIn('">folder</text>', self.curl(zs, "u2")[1])

        # fk not enabled, so this should work
        self.assertIn('">t1</text>', self.curl("dky/f.t1?th=x", "u2")[1])
        self.assertIn('">t2</text>', self.curl("dky/s1/f.t2?th=x", "u2")[1])

        ##
        ## dk+fk

        # should not be able to access dk with wrong dirkey,
        zs = self.curl("dk,fk?ls&k=%s" % (dk["dk"]), "u2")[1]
        self.assertEqual(zs, "\nJ2EOT")
        # so use the right key
        zs = self.curl("dk,fk?ls&k=%s" % (dk["dk,fk"]), "u2")[1]
        zj = json.loads(zs)
        self.assertEqual(len(zj["dirs"]), 0)
        self.assertEqual(len(zj["files"]), 1)
        self.assertEqual(zj["files"][0]["href"][:7], "f.t1?k=")

        ##
        ## dk+fk thumbs

        self.assertIn('">folder</text>', self.curl("dk,fk?th=x", "u1")[1])
        self.assertIn('">e403</text>', self.curl("dk,fk?th=x", "u2")[1])

        zs = "dk,fk?th=x&k=%s" % (dk["dk"])
        self.assertIn('">e403</text>', self.curl(zs, "u2")[1])

        zs = "dk,fk?th=x&k=%s" % (dk["dk,fk"])
        self.assertIn('">folder</text>', self.curl(zs, "u2")[1])

        # fk enabled, so this should fail
        self.assertIn('">e403</text>', self.curl("dk,fk/f.t1?th=x", "u2")[1])
        self.assertIn('">e403</text>', self.curl("dk,fk/s1/f.t2?th=x", "u2")[1])

        # but dk should return correct filekeys, so try that
        zs = "dk,fk/%s&th=x" % (zj["files"][0]["href"])
        self.assertIn('">t1</text>', self.curl(zs, "u2")[1])

        ##
        ## dks+fk

        # should not be able to access dk with wrong dirkey,
        zs = self.curl("dks,fk?ls&k=%s" % (dk["dk"]), "u2")[1]
        self.assertEqual(zs, "\nJ2EOT")
        # so use the right key
        zs = self.curl("dks,fk?ls&k=%s" % (dk["dks,fk"]), "u2")[1]
        zj = json.loads(zs)
        self.assertEqual(len(zj["dirs"]), 1)
        self.assertEqual(len(zj["files"]), 1)
        self.assertEqual(zj["dirs"][0]["href"][:6], "s1/?k=")
        self.assertEqual(zj["files"][0]["href"][:7], "f.t1?k=")

        ##
        ## dks+fk thumbs

        self.assertIn('">folder</text>', self.curl("dks,fk?th=x", "u1")[1])
        self.assertIn('">e403</text>', self.curl("dks,fk?th=x", "u2")[1])

        zs = "dks,fk?th=x&k=%s" % (dk["dk"])
        self.assertIn('">e403</text>', self.curl(zs, "u2")[1])

        zs = "dks,fk?th=x&k=%s" % (dk["dks,fk"])
        self.assertIn('">folder</text>', self.curl(zs, "u2")[1])

        # subdir s1 without key
        zs = "dks,fk/s1/?th=x"
        self.assertIn('">e403</text>', self.curl(zs, "u2")[1])

        # subdir s1 with bad key
        zs = "dks,fk/s1/?th=x&k=no"
        self.assertIn('">e403</text>', self.curl(zs, "u2")[1])

        # subdir s1 with correct key
        zs = "dks,fk/%s&th=x" % (zj["dirs"][0]["href"])
        self.assertIn('">folder</text>', self.curl(zs, "u2")[1])

        # fk enabled, so this should fail
        self.assertIn('">e403</text>', self.curl("dks,fk/f.t1?th=x", "u2")[1])
        self.assertIn('">e403</text>', self.curl("dks,fk/s1/f.t2?th=x", "u2")[1])

        # but dk should return correct filekeys, so try that
        zs = "dks,fk/%s&th=x" % (zj["files"][0]["href"])
        self.assertIn('">t1</text>', self.curl(zs, "u2")[1])

        # subdir
        self.assertIn('">e403</text>', self.curl("dks,fk/s1/?th=x", "u2")[1])
        self.assertEqual("\nJ2EOT", self.curl("dks,fk/s1/?ls", "u2")[1])
        zs = "dks,fk/s1%s&th=x" % (zj["files"][0]["href"])
        zs = self.curl("dks,fk?ls&k=%s" % (dk["dks,fk"]), "u2")[1]
        zj = json.loads(zs)
        url = "dks,fk/%s" % zj["dirs"][0]["href"]
        self.assertIn('"files"', self.curl(url + "&ls", "u2")[1])
        self.assertEqual("\nJ2EOT", self.curl(url + "x&ls", "u2")[1])

    def tardir(self, url, uname):
        top = url.split("?")[0]
        top = ("top" if not top else top.lstrip(".").split("/")[0]) + "/"
        url += ("&" if "?" in url else "?") + "tar"
        h, b = self.curl(url, uname, True)
        tar = tarfile.open(fileobj=io.BytesIO(b), mode="r|").getnames()
        if len(tar) != len([x for x in tar if x.startswith(top)]):
            raise Exception("bad-prefix:", tar)
        return " ".join([x[len(top) :] for x in tar])

    def tarsel(self, url, uname, sel):
        url += ("&" if "?" in url else "?") + "tar"
        zs = '--XD\r\nContent-Disposition: form-data; name="act"\r\n\r\nzip\r\n--XD\r\nContent-Disposition: form-data; name="files"\r\n\r\n'
        zs += "\r\n".join(sel) + "\r\n--XD--\r\n"
        zb = zs.encode("utf-8")
        hdr = "POST /%s HTTP/1.1\r\nPW: %s\r\nConnection: close\r\nContent-Type: multipart/form-data; boundary=XD\r\nContent-Length: %d\r\n\r\n"
        req = (hdr % (url, uname, len(zb))).encode("utf-8") + zb
        h, b = self.curl("", "", True, req)
        tar = tarfile.open(fileobj=io.BytesIO(b), mode="r|").getnames()
        return " ".join(tar)

    def curl(self, url, uname, binary=False, req=b""):
        req = req or hdr(url.replace("th=x", "th=j"), uname)
        conn = self.conn.setbuf(req)
        HttpCli(conn).run()
        if binary:
            h, b = conn.s._reply.split(b"\r\n\r\n", 1)
            return [h.decode("utf-8"), b]

        return conn.s._reply.decode("utf-8").split("\r\n\r\n", 1)

    def log(self, src, msg, c=0):
        print(msg, "\033[0m")



#!/usr/bin/env python

import re
import sys
import json
import time
import base64
import hashlib
import subprocess as sp

try:
    from copyparty.util import fsenc
except:

    def fsenc(p):
        return p.encode("utf-8")


"""
dep: ffmpeg
"""


def det():
    # fmt: off
    cmd = [
        b"ffmpeg",
        b"-nostdin",
        b"-hide_banner",
        b"-v", b"fatal",
        b"-i", fsenc(sys.argv[1]),
        b"-f", b"framemd5",
        b"-"
    ]
    # fmt: on

    p = sp.Popen(cmd, stdout=sp.PIPE)
    # ps = io.TextIOWrapper(p.stdout, encoding="utf-8")
    ps = p.stdout

    chans = {}
    for ln in ps:
        if ln.startswith(b"#stream#"):
            break

        m = re.match(r"^#media_type ([0-9]): ([a-zA-Z])", ln.decode("utf-8"))
        if m:
            chans[m.group(1)] = m.group(2)

    hashers = [hashlib.sha512(), hashlib.sha512()]
    for ln in ps:
        n = int(ln[:1])
        v = ln.rsplit(b",", 1)[-1].strip()
        hashers[n].update(v)

    r = {}
    for k, v in chans.items():
        dg = hashers[int(k)].digest()[:12]
        dg = base64.urlsafe_b64encode(dg).decode("ascii")
        r[v[0].lower() + "hash"] = dg

    print(json.dumps(r, indent=4))


def main():
    try:
        det()
    except:
        pass  # mute


if __name__ == "__main__":
    main()
